import numpy as np
import warnings
import sys

MAX_EXP_ARGUMENT = np.log(sys.float_info.max)


def fermi(e, mu, width=0.01):
    """
    the fermi function.
     .. math::
        f=\\frac{1}{\exp((e-\mu)/width)+1}

    :param e,mu,width: e,\mu,width
    """
    x = (e - mu) / width
    # disable overflow warning
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ret = np.where(x < MAX_EXP_ARGUMENT, 1 / (1.0 + np.exp(x)), 0.0)

    return ret



"""
This file is stolen from the hotbit programm, with some modification.
"""

import numpy as np
from ase.dft.dos import DOS
from scipy import integrate, interpolate, optimize
from scipy.optimize import brentq
from scipy.stats import logistic

# import numba

# from numba import float64, int32

# MAX_EXP_ARGUMENT = np.log(sys.float_info.max)
MAX_EXP_ARGUMENT = 700

# @numba.vectorize(nopython=True)
# def myfermi(e, mu, width, nspin):
#    x = (e - mu) / width
#    if x < -10:
#        ret = 2.0 / nspin
#    elif x > 10:
#        ret = 0.0
#    else:
#        ret = 2.0 / nspin / (math.exp(x) + 1)
#    return ret


def myfermi2(e, mu, width=0.01, nspin=1):
    """
    the fermi function.
     .. math::
        f=\\frac{1}{\exp((e-\mu)/width)+1}

    :param e,mu,width: e,\mu,width
    """
    x = (e - mu) / width
    # disable overflow warning
    # with warnings.catch_warnings():
    #    warnings.simplefilter("ignore")
    #    ret = np.where(x < MAX_EXP_ARGUMENT, (2.0 / nspin) / (1.0 + np.exp(x)), 0.0)
    xp = np.where(x < MAX_EXP_ARGUMENT, x, MAX_EXP_ARGUMENT)
    ret = (2.0 / nspin) / (1.0 + np.exp(xp))
    return ret


def myfermi(e, mu, width=0.01, nspin=1):
    return logistic.sf((e - mu) / width) * 2.0 / nspin


class Occupations(object):
    def __init__(self, nel, width, wk, nspin=1):
        """
        Initialize parameters for occupations.
        :param nel: Number of electrons
        :param width: Fermi-broadening
        :param wk: k-point weights. eg. If only gamma, [1.0]
        :param nspin(optional): number of spin, if spin=1 multiplicity=2 else, multiplicity=1.
        """
        self.nel = nel
        self.width = width
        self.wk = wk
        self.nk = len(wk)
        self.nspin = nspin
        self.mu = None

    def get_mu(self):
        """Return the Fermi-level (or chemical potential)."""
        return self.mu

    def fermi(self, mu):
        """
        Occupy states with given chemical potential.
        Occupations are 0...2; without k-point weights
        """
        return myfermi(self.e, mu, self.width, self.nspin)

    def root_function(self, mu):
        """This function is exactly zero when mu is right."""
        f = self.fermi(mu)
        return np.einsum("i, ij->", self.wk, f) - self.nel

    def occupy(self, e, xtol=1e-7):
        """
        Calculate occupation numbers with given Fermi-broadening.

        @param e: e[ind_k,ind_orb] energy of k-point, state a
            Note added by hexu:  With spin=2,e[k,a,sigma], it also work. only the *2 should be removed.
        @param wk: wk[:] weights for k-points
        @param width: The Fermi-broadening
        Returns: fermi[ind_k, ind_orb]
        """
        self.e = e
        eflat = e.flatten()
        ind = np.argsort(eflat)
        e_sorted = eflat[ind]
        if self.nspin == 1:
            m = 2
        elif self.nspin == 2:
            m = 1
        n_sorted = (self.wk[:, None, None] * np.ones_like(e) * m).flatten()[ind]

        esum = n_sorted.cumsum()
        if self.nel < esum[0]:
            ifermi = 0
        elif self.nel > esum[-1]:
            # raise ("number of electrons larger than number of orbital*spin")
            ifermi = len(e_sorted) - 1
        else:
            ifermi = np.searchsorted(esum, self.nel)
        try:
            if ifermi == 0:
                elo = e_sorted[0]
            else:
                elo = e_sorted[ifermi - 1]
            if ifermi == len(e_sorted) - 1:
                ehi = e_sorted[-1]
            else:
                ehi = e_sorted[ifermi + 1]
            guess = e_sorted[ifermi]
            dmu = np.max((self.width, guess - elo, ehi - guess))
            mu = brentq(self.root_function, guess - dmu, guess + dmu, xtol=xtol)
            # mu = brent(
            #    self.root_function,
            #    brack=(guess - elo, guess, guess + dmu),
            #    tol=xtol)
        except Exception as E:
            # probably a bad guess
            print("Error in finding Fermi level: ", E)
            dmu = self.width
            if self.nel < 1e-3:
                mu = min(e_sorted) - dmu * 20
            elif self.nel - esum[-1] > -1e-3:
                mu = max(e_sorted) + dmu * 20
            else:
                # mu = brent(
                #         self.root_function,
                #         brack=(e_sorted[0] - dmu * 10,
                #                guess,
                #                e_sorted[-1] + dmu * 10),
                #         tol=xtol)
                mu = brentq(
                    self.root_function,
                    e_sorted[0] - dmu * 20,
                    e_sorted[-1] + dmu * 20,
                    xtol=xtol,
                )

        if np.abs(self.root_function(mu)) > xtol * 1e4:
            # raise RuntimeError(
            #    'Fermi level could not be assigned reliably. Has the system fragmented?'
            # )
            print(
                "Fermi level could not be assigned reliably. Has the system fragmented?"
            )

        f = self.fermi(mu)
        # rho=(self.eigenvecs*f).dot(self.eigenvecs.transpose())

        self.mu, self.f = mu, f
        return f

    def efermi(self, e=None):
        if self.mu is None:
            self.occupy(e)
        return self.mu

    def plot(self):
        import pylab as pl

        for ik in range(self.nk):
            pl.plot(self.e[ik, :], self.f[ik, :])
            pl.scatter(self.e[ik, :], self.f[ik, :])
        pl.title("occupations")
        pl.xlabel("energy (Ha)")
        pl.ylabel("occupation")
        pl.show()


class GaussOccupations(Occupations):
    def get_mu(self):
        return self.mu

    def efermi(self, e=None):
        if self.mu is None:
            self.occupy(e)
        return self.mu

    def delta(self, energy):
        """Return a delta-function centered at 'energy'."""
        x = -(((self.e - energy) / self.width) ** 2)
        return np.exp(x) / (np.sqrt(np.pi) * self.width)

    def get_dos(self, npts=500):
        dos = np.zeros(npts)
        for w, e_n in zip(self.w_k, self.e_skn[0]):
            for e in e_n:
                dos += w * self.delta(e)

    def root_function(self, mu):
        pass

    # @profile
    def occupy(self, e, xtol=1e-8, guess=0.0, npts=2001):
        self.e = e
        dos = myDOS(kweights=self.wk, eigenvalues=e, width=self.width, npts=npts)
        edos = dos.get_energies()
        d = dos.get_dos()
        #idos = integrate.cumtrapz(d, edos, initial=0) - self.nel
        idos = integrate.cumulative_trapezoid(d, edos, initial=0) - self.nel
        # ifermi = np.searchsorted(idos, 0.0)
        # self.mu = edos[ifermi]

        # interpolate the idos and find the root
        f_idos = interpolate.interp1d(edos, idos, kind="cubic")
        # ret = optimize.fmin(f_idos, x0=edos[40], xtol=xtol, disp=True, )
        # find the 0 value of f_idos
        ret = optimize.root_scalar(
            f_idos,
            x0=edos[npts // 2],
            xtol=xtol,
            method="brentq",
            bracket=[edos[0], edos[-1]],
        )
        self.mu = ret.root
        if not ret.converged:
            print("Warning: Fermi level not converged")
        self.f = self.fermi(self.mu)
        return self.f


class myDOS(DOS):
    def __init__(
        self, kweights, eigenvalues, nspin=1, width=0.1, window=None, npts=1001
    ):
        """Electronic Density Of States object.

        calc: calculator object
            Any ASE compliant calculator object.
        width: float
            Width of guassian smearing.  Use width=0.0 for linear tetrahedron
            interpolation.
        window: tuple of two float
            Use ``window=(emin, emax)``.  If not specified, a window
            big enough to hold all the eigenvalues will be used.
        npts: int
            Number of points.

        """
        self.npts = npts
        self.width = width
        # self.w_k = calc.get_k_point_weights()
        self.w_k = kweights
        self.nspins = nspin
        # self.e_skn = np.array([[calc.get_eigenvalues(kpt=k, spin=s)
        #                        for k in range(len(self.w_k))]
        #                       for s in range(self.nspins)])
        # self.e_skn -= calc.get_fermi_level()
        self.e_skn = np.array([eigenvalues])  # eigenvalues: iband, ikpt

        if window is None:
            emin = None
            emax = None
        else:
            emin, emax = window

        if emin is None:
            emin = self.e_skn.min() - 10 * self.width
        if emax is None:
            emax = self.e_skn.max() + 10 * self.width

        self.energies = np.linspace(emin, emax, npts)

        # if width == 0.0: # To use tetrahedron method
        #    bzkpts = calc.get_bz_k_points()
        #    size, offset = get_monkhorst_pack_size_and_offset(bzkpts)
        #    bz2ibz = calc.get_bz_to_ibz_map()
        #    shape = (self.nspins,) + tuple(size) + (-1,)
        #    self.e_skn = self.e_skn[:, bz2ibz].reshape(shape)
        #    self.cell = calc.atoms.cell

    def get_idos(self):
        e, d = self.get_dos()
        return np.trapz(d, e)

    def delta(self, energy):
        """Return a delta-function centered at 'energy'."""
        x = -(((self.energies - energy) / self.width) ** 2)
        return np.exp(x) / (np.sqrt(np.pi) * self.width)
        # use error function to avoid overflow
        # x = (self.energies - energy) / self.width
        # return 0.5 * (np.erf(x) - np.erf(-x))

    def get_dos(self, spin=None):
        """Get array of DOS values.

        The *spin* argument can be 0 or 1 (spin up or down) - if not
        specified, the total DOS is returned.
        """

        if spin is None:
            if self.nspins == 2:
                # Spin-polarized calculation, but no spin specified -
                # return the total DOS:
                return self.get_dos(spin=0) + self.get_dos(spin=1)
            else:
                spin = 0

        dos = np.zeros(self.npts)
        for w, e_n in zip(self.w_k, self.e_skn[spin]):
            for e in e_n:
                dos += w * self.delta(e)
        return dos



"""Plotting utilities for EPW data."""

import os
from .epwparser import Epmat, read_crystal_fmt
from ase.units import Bohr


def plot_epw_distance(
    ax,
    path,
    imode,
    distance_type="Rk",
    epmat_ncfile="epmat.nc",
    crystal_fmt_file="crystal.fmt",
    ylim=(1e-5, 10),
    **scatter_kwargs,
):
    """
    Plot electron-phonon coupling matrix elements vs. distance.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The matplotlib axes object to plot on.
    path : str
        Directory containing the EPW files.
    imode : int
        The phonon mode index to analyze.
    distance_type : str, optional
        The type of distance to plot, either 'Rk' (WF-WF) or 'Rg' (WF-atom).
        Default is 'Rk'.
    epmat_ncfile : str, optional
        Name of the NetCDF file. Default is 'epmat.nc'.
    crystal_fmt_file : str, optional
        Name of the crystal format file. Default is 'crystal.fmt'.
    ylim : tuple, optional
        The y-axis limits for the plot. Default is (1e-5, 10).
    **scatter_kwargs : dict
        Additional keyword arguments to pass to `ax.scatter`.
    """
    # Load data
    crystal = read_crystal_fmt(os.path.join(path, crystal_fmt_file))
    ep = Epmat()
    ep.crystal = crystal
    ep.read(
        path=path, prefix="prefix", epmat_ncfile=epmat_ncfile
    )  # prefix is not used when ncfile is present

    # Get cell
    cell = crystal.at.reshape(3, 3) * crystal.alat * Bohr

    # Calculate distances
    if distance_type == "Rk":
        entries = ep.distance_resolved_couplings_Rk(imode=imode, cell=cell)
    elif distance_type == "Rg":
        entries = ep.distance_resolved_couplings_Rg(imode=imode, cell=cell)
    else:
        raise ValueError("distance_type must be 'Rk' or 'Rg'")

    distances = [e["distance"] for e in entries]
    magnitudes = [abs(e["coupling"]) for e in entries]

    # Default scatter settings
    defaults = {"s": 5, "alpha": 0.6}
    defaults.update(scatter_kwargs)

    # Plot
    ax.scatter(distances, magnitudes, **defaults)
    ax.set_xlabel("Distance (Å)")
    ax.set_ylabel(r"|$g$| (eV/Å)")
    ax.set_yscale("log")
    ax.set_ylim(ylim)

    return ax



"""
Wigner-Seitz data structure for EPW calculations.

This module provides a dataclass to handle Wigner-Seitz data for electrons,
phonons, and electron-phonon interactions as used in EPW calculations.
"""

from dataclasses import dataclass
from typing import List, Tuple
import numpy as np


@dataclass
class WignerData:
    """
    Wigner-Seitz data container for EPW calculations.

    This dataclass stores the Wigner-Seitz vectors, degeneracies, and lengths
    for electrons (k), phonons (q), and electron-phonon (g) interactions.

    Attributes:
        dims: Number of Wannier functions (if use_ws, otherwise 1)
        dims2: Number of atoms (if use_ws, otherwise 1)

        # Electron data (k)
        nrr_k: Number of WS vectors for electrons
        irvec_k: WS vectors for electrons, shape (nrr_k, 3)
        ndegen_k: WS degeneracies for electrons, shape (nrr_k, dims, dims)
        wslen_k: WS vector lengths for electrons, shape (nrr_k,)

        # Phonon data (q)
        nrr_q: Number of WS vectors for phonons
        irvec_q: WS vectors for phonons, shape (nrr_q, 3)
        ndegen_q: WS degeneracies for phonons, shape (nrr_q, dims2, dims2)
        wslen_q: WS vector lengths for phonons, shape (nrr_q,)

        # Electron-phonon data (g)
        nrr_g: Number of WS vectors for electron-phonon
        irvec_g: WS vectors for electron-phonon, shape (nrr_g, 3)
        ndegen_g: WS degeneracies for electron-phonon, shape (nrr_g, dims, dims2)
        wslen_g: WS vector lengths for electron-phonon, shape (nrr_g,)
    """

    # Dimensions
    dims: int
    dims2: int

    # Electron data (k)
    nrr_k: int
    irvec_k: np.ndarray  # shape (nrr_k, 3)
    ndegen_k: np.ndarray  # shape (nrr_k, dims, dims)
    wslen_k: np.ndarray  # shape (nrr_k,)

    # Phonon data (q)
    nrr_q: int
    irvec_q: np.ndarray  # shape (nrr_q, 3)
    ndegen_q: np.ndarray  # shape (nrr_q, dims2, dims2)
    wslen_q: np.ndarray  # shape (nrr_q,)

    # Electron-phonon data (g)
    nrr_g: int
    irvec_g: np.ndarray  # shape (nrr_g, 3)
    ndegen_g: np.ndarray  # shape (nrr_g, dims, dims2)
    wslen_g: np.ndarray  # shape (nrr_g,)

    def __post_init__(self):
        """Validate array shapes after initialization."""
        # Convert to numpy arrays if needed
        self.irvec_k = np.asarray(self.irvec_k)
        self.ndegen_k = np.asarray(self.ndegen_k)
        self.wslen_k = np.asarray(self.wslen_k)

        self.irvec_q = np.asarray(self.irvec_q)
        self.ndegen_q = np.asarray(self.ndegen_q)
        self.wslen_q = np.asarray(self.wslen_q)

        self.irvec_g = np.asarray(self.irvec_g)
        self.ndegen_g = np.asarray(self.ndegen_g)
        self.wslen_g = np.asarray(self.wslen_g)

        # Validate shapes
        assert self.irvec_k.shape == (
            self.nrr_k,
            3,
        ), f"irvec_k shape mismatch: {self.irvec_k.shape} != ({self.nrr_k}, 3)"
        assert self.ndegen_k.shape == (
            self.nrr_k,
            self.dims,
            self.dims,
        ), f"ndegen_k shape mismatch"
        assert self.wslen_k.shape == (self.nrr_k,), f"wslen_k shape mismatch"

        assert self.irvec_q.shape == (self.nrr_q, 3), f"irvec_q shape mismatch"
        assert self.ndegen_q.shape == (
            self.nrr_q,
            self.dims2,
            self.dims2,
        ), f"ndegen_q shape mismatch"
        assert self.wslen_q.shape == (self.nrr_q,), f"wslen_q shape mismatch"

        assert self.irvec_g.shape == (self.nrr_g, 3), f"irvec_g shape mismatch"
        assert self.ndegen_g.shape == (
            self.nrr_g,
            self.dims,
            self.dims2,
        ), f"ndegen_g shape mismatch"
        assert self.wslen_g.shape == (self.nrr_g,), f"wslen_g shape mismatch"

    @classmethod
    def from_file(cls, filename: str = "wigner.fmt") -> "WignerData":
        """
        Read Wigner-Seitz data from file format written by EPW.

        Args:
            filename: Path to the wigner data file

        Returns:
            WignerData instance
        """
        with open(filename, "r") as f:
            # Read header
            header = f.readline().strip().split()
            nrr_k, nrr_q, nrr_g, dims, dims2 = map(int, header)

            # Initialize arrays
            irvec_k = np.zeros((nrr_k, 3), dtype=int)
            wslen_k = np.zeros(nrr_k, dtype=float)
            ndegen_k = np.zeros((nrr_k, dims, dims), dtype=int)

            irvec_q = np.zeros((nrr_q, 3), dtype=int)
            wslen_q = np.zeros(nrr_q, dtype=float)
            ndegen_q = np.zeros((nrr_q, dims2, dims2), dtype=int)

            irvec_g = np.zeros((nrr_g, 3), dtype=int)
            wslen_g = np.zeros(nrr_g, dtype=float)
            ndegen_g = np.zeros((nrr_g, dims, dims2), dtype=int)

            # Read electron data (k)
            for ir in range(nrr_k):
                line = f.readline().strip().split()
                irvec_k[ir, :] = [int(line[0]), int(line[1]), int(line[2])]
                wslen_k[ir] = float(line[3])

                for iw in range(dims):
                    line = f.readline().strip().split()
                    ndegen_k[ir, iw, :] = [int(x) for x in line]

            # Read phonon data (q)
            for ir in range(nrr_q):
                line = f.readline().strip().split()
                irvec_q[ir, :] = [int(line[0]), int(line[1]), int(line[2])]
                wslen_q[ir] = float(line[3])

                for na in range(dims2):
                    line = f.readline().strip().split()
                    ndegen_q[ir, na, :] = [int(x) for x in line]

            # Read electron-phonon data (g)
            for ir in range(nrr_g):
                line = f.readline().strip().split()
                irvec_g[ir, :] = [int(line[0]), int(line[1]), int(line[2])]
                wslen_g[ir] = float(line[3])

                for iw in range(dims):
                    line = f.readline().strip().split()
                    ndegen_g[ir, iw, :] = [int(x) for x in line]

        return cls(
            dims=dims,
            dims2=dims2,
            nrr_k=nrr_k,
            irvec_k=irvec_k,
            ndegen_k=ndegen_k,
            wslen_k=wslen_k,
            nrr_q=nrr_q,
            irvec_q=irvec_q,
            ndegen_q=ndegen_q,
            wslen_q=wslen_q,
            nrr_g=nrr_g,
            irvec_g=irvec_g,
            ndegen_g=ndegen_g,
            wslen_g=wslen_g,
        )

    def to_file(self, filename: str = "wigner.fmt") -> None:
        """
        Write Wigner-Seitz data to file in EPW format.

        Args:
            filename: Output file path
        """
        with open(filename, "w") as f:
            # Write header
            f.write(
                f"{self.nrr_k} {self.nrr_q} {self.nrr_g} {self.dims} {self.dims2}\n"
            )

            # Write electron data (k)
            for ir in range(self.nrr_k):
                f.write(
                    f"{self.irvec_k[ir, 0]:6d}{self.irvec_k[ir, 1]:6d}{self.irvec_k[ir, 2]:6d}"
                )
                f.write(f"{self.wslen_k[ir]:26.17E}\n")

                for iw in range(self.dims):
                    degeneracies = " ".join(
                        str(self.ndegen_k[ir, iw, j]) for j in range(self.dims)
                    )
                    f.write(f"{degeneracies}\n")

            # Write phonon data (q)
            for ir in range(self.nrr_q):
                f.write(
                    f"{self.irvec_q[ir, 0]:6d}{self.irvec_q[ir, 1]:6d}{self.irvec_q[ir, 2]:6d}"
                )
                f.write(f"{self.wslen_q[ir]:26.17E}\n")

                for na in range(self.dims2):
                    degeneracies = " ".join(
                        str(self.ndegen_q[ir, na, j]) for j in range(self.dims2)
                    )
                    f.write(f"{degeneracies}\n")

            # Write electron-phonon data (g)
            for ir in range(self.nrr_g):
                f.write(
                    f"{self.irvec_g[ir, 0]:6d}{self.irvec_g[ir, 1]:6d}{self.irvec_g[ir, 2]:6d}"
                )
                f.write(f"{self.wslen_g[ir]:26.17E}\n")

                for iw in range(self.dims):
                    degeneracies = " ".join(
                        str(self.ndegen_g[ir, iw, j]) for j in range(self.dims2)
                    )
                    f.write(f"{degeneracies}\n")

    def summary(self) -> str:
        """Return a summary string of the Wigner data."""
        return (
            f"WignerData Summary:\n"
            f"  Dimensions: {self.dims} Wannier functions, {self.dims2} atoms\n"
            f"  Electrons (k): {self.nrr_k} WS vectors\n"
            f"  Phonons (q): {self.nrr_q} WS vectors\n"
            f"  Electron-phonon (g): {self.nrr_g} WS vectors"
        )

    def find_origin_indices(self) -> Tuple[int, int, int]:
        """
        Find the indices of the origin vectors (0,0,0) for k, q, and g.

        Returns:
            Tuple of (k_origin_idx, q_origin_idx, g_origin_idx)
            Returns -1 if origin vector not found
        """
        k_origin = np.where(np.all(self.irvec_k == 0, axis=1))[0]
        q_origin = np.where(np.all(self.irvec_q == 0, axis=1))[0]
        g_origin = np.where(np.all(self.irvec_g == 0, axis=1))[0]

        return (
            k_origin[0] if len(k_origin) > 0 else -1,
            q_origin[0] if len(q_origin) > 0 else -1,
            g_origin[0] if len(g_origin) > 0 else -1,
        )




from HamiltonIO.hamiltonian import Hamiltonian
from HamiltonIO.siesta import SiestaHam
from HamiltonIO.wannier import WannierHam

__version__ = "0.2.5"
__all__ = ["Hamiltonian", "SiestaHam", "WannierHam", "__version__"]



#!/usr/bin/env python3
"""
HamiltonIO EPW Command Line Interface

Provides command-line tools for converting and analyzing EPW
(Electron-Phonon Wannier) data.
"""

import argparse
import sys
import time
from pathlib import Path

from .epwparser import save_epmat_to_nc, Epmat, read_crystal_fmt
from .plot import plot_epw_distance


def get_version():
    """Get the HamiltonIO version."""
    try:
        from .. import __version__

        return __version__
    except ImportError:
        return "0.2.5"  # Fallback version from pyproject.toml


def validate_epw_files(path, prefix):
    """Validate that all required EPW files exist and return their info."""
    required_files = ["epwdata.fmt", "wigner.fmt", f"{prefix}.epmatwp"]
    file_info = []
    missing_files = []

    for filename in required_files:
        filepath = Path(path) / filename
        if filepath.exists():
            size = filepath.stat().st_size
            file_info.append((filename, size))
        else:
            missing_files.append(filename)

    return file_info, missing_files


def format_file_size(size_bytes):
    """Format file size in human-readable format."""
    if size_bytes < 1024:
        return f"{size_bytes} B"
    elif size_bytes < 1024**2:
        return f"{size_bytes / 1024:.1f} KB"
    elif size_bytes < 1024**3:
        return f"{size_bytes / (1024**2):.1f} MB"
    else:
        return f"{size_bytes / (1024**3):.1f} GB"


def convert_to_netcdf(args):
    """Convert EPW binary files to NetCDF format."""
    print("=== HamiltonIO EPW to NetCDF Converter ===")
    print(f"Version: {get_version()}")
    print()

    # Validate input directory
    input_path = Path(args.path)
    if not input_path.exists():
        print(f" Error: Directory not found: {input_path}")
        sys.exit(1)

    if not input_path.is_dir():
        print(f" Error: Path is not a directory: {input_path}")
        sys.exit(1)

    print(f" Input directory: {input_path.absolute()}")
    print(f" EPW file prefix: {args.prefix}")
    print(f" Output file: {args.output}")
    if args.dry_run:
        print(" Mode: Dry run (no conversion will be performed)")
    print()

    # Validate required files
    print(" Checking required files...")
    file_info, missing_files = validate_epw_files(args.path, args.prefix)

    total_size = 0
    for filename, size in file_info:
        print(f" {filename}: {format_file_size(size)}")
        total_size += size

    print(f"\n Total input size: {format_file_size(total_size)}")

    if missing_files:
        print("\n Missing required files:")
        for filename in missing_files:
            print(f"   - {filename}")
        print("\nPlease ensure all required EPW output files are present.")
        sys.exit(1)

    # Check output file
    output_path = Path(args.path) / args.output
    if output_path.exists():
        print(f"\n  Warning: Output file exists: {output_path}")
        if not args.force:
            response = input("Overwrite? (y/N): ")
            if response.lower() != "y":
                print("Conversion cancelled.")
                sys.exit(0)
        else:
            print("Overwriting existing file (force mode).")

    # Dry run mode - just check files and exit
    if args.dry_run:
        print("\n Dry run completed - all required files are present and valid.")
        print("   Run without --dry-run to perform the actual conversion.")
        return

    # Perform conversion
    print(f"\n Converting {args.prefix}.epmatwp to {args.output}...")
    print("   This may take a while for large files...")

    start_time = time.time()
    try:
        save_epmat_to_nc(path=args.path, prefix=args.prefix, ncfile=args.output)
        conversion_time = time.time() - start_time

        if output_path.exists():
            output_size = output_path.stat().st_size
            print(" Conversion completed successfully!")
            print(f"   Output file: {output_path.absolute()}")
            print(f"   File size: {format_file_size(output_size)}")
            print(f"   Conversion time: {conversion_time:.2f} seconds")
        else:
            print(" Error: Conversion failed - output file not created")
            sys.exit(1)

    except KeyboardInterrupt:
        print("\n Conversion cancelled by user")
        sys.exit(1)
    except Exception as e:
        print("\n Error: Conversion failed")
        print(f"   Details: {str(e)}")
        sys.exit(1)


def distance_analysis(args):
    """Analyze electron-phonon coupling vs. distance."""
    import matplotlib

    matplotlib.use("Agg")  # Non-interactive backend
    import matplotlib.pyplot as plt
    from ase.units import Bohr

    print("=== HamiltonIO EPW Distance Analysis ===")
    print(f"Version: {get_version()}")
    print()

    # Validate input directory
    input_path = Path(args.path)
    if not input_path.exists():
        print(f" Error: Directory not found: {input_path}")
        sys.exit(1)

    print(f" Input directory: {input_path.absolute()}")
    print(f" NetCDF file: {args.epmat_ncfile}")
    print(f" Crystal file: {args.crystal_fmt_file}")
    print(f" Mode index: {args.imode}")
    print(f" Distance type: {args.distance_type}")
    print(f" Output file: {args.output}")
    print()

    # Check required files
    ncfile_path = input_path / args.epmat_ncfile
    crystal_path = input_path / args.crystal_fmt_file

    if not ncfile_path.exists():
        print(f" Error: NetCDF file not found: {ncfile_path}")
        print(" Run 'hamiltonio-epw epw_to_nc' first to convert EPW data.")
        sys.exit(1)

    if not crystal_path.exists():
        print(f" Error: Crystal file not found: {crystal_path}")
        sys.exit(1)

    # Load data
    print(" Loading EPW data...")
    try:
        crystal = read_crystal_fmt(str(crystal_path))
        ep = Epmat()
        ep.crystal = crystal
        ep.read(path=str(input_path), prefix="prefix", epmat_ncfile=args.epmat_ncfile)
    except Exception as e:
        print(f" Error loading data: {e}")
        sys.exit(1)

    # Get cell
    cell = crystal.at.reshape(3, 3) * crystal.alat * Bohr

    # Calculate distances
    print(f" Calculating {args.distance_type}-based distances for mode {args.imode}...")
    try:
        if args.distance_type == "Rk":
            entries = ep.distance_resolved_couplings_Rk(imode=args.imode, cell=cell)
        elif args.distance_type == "Rg":
            entries = ep.distance_resolved_couplings_Rg(imode=args.imode, cell=cell)
        else:
            print(
                f" Error: Invalid distance_type '{args.distance_type}'."
                " Use 'Rk' or 'Rg'."
            )
            sys.exit(1)
    except Exception as e:
        print(f" Error calculating distances: {e}")
        sys.exit(1)

    print(f" Found {len(entries)} non-zero coupling elements")

    # Create plot
    print(" Creating plot...")
    fig, ax = plt.subplots(figsize=(6, 5))

    try:
        plot_epw_distance(
            ax=ax,
            path=str(input_path),
            imode=args.imode,
            distance_type=args.distance_type,
            epmat_ncfile=args.epmat_ncfile,
            crystal_fmt_file=args.crystal_fmt_file,
            ylim=args.ylim,
        )

        # Add title
        distance_label = "WF-WF" if args.distance_type == "Rk" else "WF-atom"
        ax.set_title(f"Mode {args.imode}: {distance_label} distance")

        plt.tight_layout()
        plt.savefig(args.output, dpi=150)

        output_size = Path(args.output).stat().st_size
        print(" Plot created successfully!")
        print(f"   Output file: {Path(args.output).absolute()}")
        print(f"   File size: {format_file_size(output_size)}")

    except Exception as e:
        print(f" Error creating plot: {e}")
        sys.exit(1)


def main():
    """Main CLI entry point."""
    parser = argparse.ArgumentParser(
        prog="hamiltonio-epw",
        description="HamiltonIO EPW (Electron-Phonon Wannier) tools",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  # Convert EPW data to NetCDF
  %(prog)s epw_to_nc --path ./epw_data --prefix material --output epmat.nc
  %(prog)s epw_to_nc -p /path/to/files -n my_material -o output.nc
  %(prog)s epw_to_nc --path ./ --prefix test --force

  # Analyze distance-resolved coupling
  %(prog)s distance --path ./epw_data --imode 0 --distance-type Rk
  %(prog)s distance -p ./ -m 5 -t Rg -o mode5_Rg.pdf
  %(prog)s distance -m 0 --ylim 1e-4 10

For more information, visit: https://github.com/mailhexu/HamiltonIO
        """,
    )

    parser.add_argument(
        "--version", action="version", version=f"HamiltonIO EPW CLI {get_version()}"
    )

    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Convert command
    convert_parser = subparsers.add_parser(
        "epw_to_nc", help="Convert EPW binary files to NetCDF format"
    )
    convert_parser.add_argument(
        "-p",
        "--path",
        default=".",
        help="Path to directory containing EPW files (default: current directory)",
    )
    convert_parser.add_argument(
        "-n", "--prefix", default="epw", help="EPW file prefix (default: epw)"
    )
    convert_parser.add_argument(
        "-o",
        "--output",
        default="epmat.nc",
        help="Output NetCDF filename (default: epmat.nc)",
    )
    convert_parser.add_argument(
        "--force",
        action="store_true",
        help="Overwrite existing output file without prompting",
    )
    convert_parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Check files and show conversion info without performing conversion",
    )

    # Distance analysis command
    distance_parser = subparsers.add_parser(
        "distance",
        help="Analyze electron-phonon coupling vs. distance",
    )
    distance_parser.add_argument(
        "-p",
        "--path",
        default=".",
        help="Path to directory containing EPW files (default: current directory)",
    )
    distance_parser.add_argument(
        "-m",
        "--imode",
        type=int,
        required=True,
        help="Phonon mode index to analyze",
    )
    distance_parser.add_argument(
        "-t",
        "--distance-type",
        choices=["Rk", "Rg"],
        default="Rk",
        help="Distance type: 'Rk' for WF-WF, 'Rg' for WF-atom (default: Rk)",
    )
    distance_parser.add_argument(
        "-o",
        "--output",
        default="epw_distance.pdf",
        help="Output plot filename (default: epw_distance.pdf)",
    )
    distance_parser.add_argument(
        "--epmat-ncfile",
        default="epmat.nc",
        help="NetCDF file containing EPW data (default: epmat.nc)",
    )
    distance_parser.add_argument(
        "--crystal-fmt-file",
        default="crystal.fmt",
        help="Crystal format file (default: crystal.fmt)",
    )
    distance_parser.add_argument(
        "--ylim",
        nargs=2,
        type=float,
        default=[1e-5, 10],
        metavar=("MIN", "MAX"),
        help="Y-axis limits (default: 1e-5 10)",
    )

    # Parse arguments
    args = parser.parse_args()

    # Handle commands
    if args.command == "epw_to_nc":
        convert_to_netcdf(args)
    elif args.command == "distance":
        distance_analysis(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Unit tests for SIESTA CLI functionality.

Purpose:
    Test the command-line interface and programmatic API for SIESTA intra-atomic analysis.

How to run:
    uv run python -m pytest tests/test_siesta_cli.py -v
"""

import os
import tempfile
import pytest
from pathlib import Path


@pytest.fixture
def bcc_fe_soc_path():
    """Path to BCC Fe SOC example."""
    base = Path(__file__).parent.parent
    fdf_path = base / "examples" / "SIESTA" / "bccFe" / "SOC" / "siesta.fdf"
    if not fdf_path.exists():
        pytest.skip("BCC Fe SOC example data not found")
    return str(fdf_path)


@pytest.fixture
def bcc_fe_nonpol_path():
    """Path to BCC Fe nonpolarized example."""
    base = Path(__file__).parent.parent
    fdf_path = base / "examples" / "SIESTA" / "bccFe" / "nonpolarized" / "siesta.fdf"
    if not fdf_path.exists():
        pytest.skip("BCC Fe nonpolarized example data not found")
    return str(fdf_path)


def test_analyze_intra_atomic_function_basic(bcc_fe_nonpol_path):
    """Test basic usage of analyze_intra_atomic function."""
    from HamiltonIO.siesta.cli import analyze_intra_atomic

    # Analyze without output file (should not raise error)
    ham = analyze_intra_atomic(
        fdf_file=bcc_fe_nonpol_path,
        atoms=None,
        read_soc=False,
        pauli_decomp=True,
        show_matrix=False,
        ispin="merge",
        output_file=None,
    )

    # Check returned Hamiltonian
    assert ham is not None
    assert ham.nbasis > 0
    assert len(ham.atoms) > 0


def test_analyze_intra_atomic_function_with_output(bcc_fe_nonpol_path):
    """Test analyze_intra_atomic with output file."""
    from HamiltonIO.siesta.cli import analyze_intra_atomic

    with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".txt") as f:
        output_file = f.name

    try:
        ham = analyze_intra_atomic(
            fdf_file=bcc_fe_nonpol_path,
            atoms=[0],
            read_soc=False,
            pauli_decomp=True,
            show_matrix=True,
            ispin="merge",
            output_file=output_file,
        )

        # Check file was created
        assert os.path.exists(output_file)
        assert os.path.getsize(output_file) > 0

        # Check content
        with open(output_file, "r") as f:
            content = f.read()
            assert "Intra-Atomic Hamiltonian Analysis" in content
            assert "Atom 0" in content
            assert "Fe" in content

    finally:
        if os.path.exists(output_file):
            os.remove(output_file)


def test_analyze_intra_atomic_function_atom_filtering(bcc_fe_nonpol_path):
    """Test analyze_intra_atomic with atom filtering."""
    from HamiltonIO.siesta.cli import analyze_intra_atomic

    with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".txt") as f:
        output_file = f.name

    try:
        ham = analyze_intra_atomic(
            fdf_file=bcc_fe_nonpol_path,
            atoms=[0],  # Only first atom
            read_soc=False,
            pauli_decomp=False,
            show_matrix=False,
            ispin="merge",
            output_file=output_file,
        )

        # Check output contains only atom 0
        with open(output_file, "r") as f:
            content = f.read()
            assert "Atom 0" in content
            # Should not have Atom 1 if there are multiple atoms
            if len(ham.atoms) > 1:
                assert "Atom 1" not in content

    finally:
        if os.path.exists(output_file):
            os.remove(output_file)


def test_analyze_intra_atomic_soc(bcc_fe_soc_path):
    """Test analyze_intra_atomic with SOC system."""
    from HamiltonIO.siesta.cli import analyze_intra_atomic

    with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".txt") as f:
        output_file = f.name

    try:
        ham = analyze_intra_atomic(
            fdf_file=bcc_fe_soc_path,
            atoms=[0],
            read_soc=False,
            pauli_decomp=True,
            show_matrix=True,
            ispin="merge",
            output_file=output_file,
        )

        # Check content has Pauli decomposition
        with open(output_file, "r") as f:
            content = f.read()
            assert "Pauli Decomposition" in content
            # Should have spin components
            assert "spin" in content.lower() or "σ" in content

    finally:
        if os.path.exists(output_file):
            os.remove(output_file)


def test_cli_main_function_help():
    """Test CLI main function with --help."""
    from HamiltonIO.siesta.cli import main
    import sys

    # Save original argv
    original_argv = sys.argv

    try:
        # Test --help
        sys.argv = ["hamiltonio-siesta", "--help"]
        with pytest.raises(SystemExit) as excinfo:
            main()
        # --help should exit with code 0
        assert excinfo.value.code == 0

    finally:
        # Restore original argv
        sys.argv = original_argv


def test_cli_intra_atomic_command_help():
    """Test CLI intra-atomic subcommand with --help."""
    from HamiltonIO.siesta.cli import main
    import sys

    original_argv = sys.argv

    try:
        sys.argv = ["hamiltonio-siesta", "intra-atomic", "--help"]
        with pytest.raises(SystemExit) as excinfo:
            main()
        assert excinfo.value.code == 0

    finally:
        sys.argv = original_argv


def test_cli_version():
    """Test CLI --version flag."""
    from HamiltonIO.siesta.cli import main
    import sys

    original_argv = sys.argv

    try:
        sys.argv = ["hamiltonio-siesta", "--version"]
        with pytest.raises(SystemExit) as excinfo:
            main()
        # Version should exit with code 0
        assert excinfo.value.code == 0

    finally:
        sys.argv = original_argv


def test_cli_intra_atomic_command_basic(bcc_fe_nonpol_path):
    """Test CLI intra-atomic command basic usage."""
    from HamiltonIO.siesta.cli import main
    import sys

    with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".txt") as f:
        output_file = f.name

    original_argv = sys.argv

    try:
        sys.argv = [
            "hamiltonio-siesta",
            "intra-atomic",
            bcc_fe_nonpol_path,
            "-o",
            output_file,
            "--atoms",
            "0",
        ]
        main()

        # Check output file
        assert os.path.exists(output_file)
        with open(output_file, "r") as f:
            content = f.read()
            assert "Atom 0" in content

    finally:
        sys.argv = original_argv
        if os.path.exists(output_file):
            os.remove(output_file)


def test_cli_invalid_fdf_file():
    """Test CLI with invalid FDF file."""
    from HamiltonIO.siesta.cli import main
    import sys

    original_argv = sys.argv

    try:
        sys.argv = [
            "hamiltonio-siesta",
            "intra-atomic",
            "/nonexistent/path/siesta.fdf",
        ]
        with pytest.raises(SystemExit):
            main()

    finally:
        sys.argv = original_argv


if __name__ == "__main__":
    # Run tests
    pytest.main([__file__, "-v"])



"""Tests for distance-resolved hopping analysis in Wannier90 Hamiltonian.

Tests cover:
    - Correctness of iter_unique_hoppings (symmetry reduction)
    - Distance computation from Wannier centers
    - Binning of hopping magnitudes by distance
"""

import os
from pathlib import Path

import numpy as np
import pytest
from ase.io import read

from HamiltonIO.wannier import WannierHam


@pytest.fixture
def srmo3_ham():
    """Load SrMnO3 Wannier TB model from test data."""
    test_dir = Path(__file__).parent / "data" / "SrMnO3_wannier"
    pwi_file = test_dir / "SrMnO3.pwi"
    atoms = read(pwi_file, format="espresso-in")
    ham = WannierHam.read_from_wannier_dir(test_dir, "SrMnO3", atoms=atoms)
    return ham, atoms


def test_iter_unique_hoppings(srmo3_ham):
    """Test that iter_unique_hoppings yields unique hopping terms."""
    ham, _ = srmo3_ham

    # Collect all unique hoppings
    unique_hops = list(ham.iter_unique_hoppings())

    # Should have at least some hoppings
    assert len(unique_hops) > 0

    # Check structure: (R, i, j, value)
    for R, i, j, val in unique_hops:
        assert len(R) == 3, "R should be 3D"
        assert isinstance(i, (int, np.integer)), "i should be integer"
        assert isinstance(j, (int, np.integer)), "j should be integer"
        assert isinstance(val, (complex, np.complexfloating)), "value should be complex"

    # Check symmetry convention: only positive R (first non-zero component >= 0)
    for R, i, j, val in unique_hops:
        nz = np.nonzero(R)[0]
        if len(nz) > 0:
            assert R[nz[0]] >= 0, (
                f"First non-zero R component should be >= 0, got R={R}"
            )

    # Check R=0 convention: only upper triangle (i <= j)
    r0_hops = [(i, j) for R, i, j, _ in unique_hops if np.all(np.array(R) == 0)]
    for i, j in r0_hops:
        assert i <= j, f"For R=0, should have i <= j, got i={i}, j={j}"


def test_distance_resolved_hoppings(srmo3_ham):
    """Test distance computation for hopping matrix elements."""
    ham, atoms = srmo3_ham
    cell = atoms.get_cell()

    # Get distance-resolved hoppings
    entries = ham.distance_resolved_hoppings(cell)

    # Should have hoppings
    assert len(entries) > 0

    # Check structure of each entry
    for entry in entries:
        assert "R" in entry
        assert "i" in entry
        assert "j" in entry
        assert "distance" in entry
        assert "hopping" in entry

        # Check types
        assert len(entry["R"]) == 3
        assert isinstance(entry["i"], int)
        assert isinstance(entry["j"], int)
        assert isinstance(entry["distance"], float)
        assert entry["distance"] >= 0, "Distance should be non-negative"
        assert isinstance(entry["hopping"], complex)

    # Distances should be reasonable (not all zero, not all huge)
    distances = np.array([e["distance"] for e in entries])
    assert distances.min() >= 0
    assert distances.max() < 100, "Distances should be reasonable (< 100 Å)"
    assert len(np.unique(distances)) > 1, "Should have multiple distinct distances"


def test_distance_resolved_hoppings_onsite(srmo3_ham):
    """Test that R=0, i=j hoppings have near-zero distance."""
    ham, atoms = srmo3_ham
    cell = atoms.get_cell()

    entries = ham.distance_resolved_hoppings(cell)

    # Find onsite hoppings (R=0, i=j)
    onsite = [e for e in entries if e["R"] == (0, 0, 0) and e["i"] == e["j"]]

    # Should have some onsite terms
    assert len(onsite) > 0, "Should have onsite diagonal elements"

    # Onsite distances should be very small (essentially zero)
    for entry in onsite:
        assert entry["distance"] < 1e-6, (
            f"Onsite distance should be ~0, got {entry['distance']}"
        )


def test_bin_hoppings_by_distance(srmo3_ham):
    """Test binning of hopping magnitudes by distance."""
    ham, atoms = srmo3_ham
    cell = atoms.get_cell()

    entries = ham.distance_resolved_hoppings(cell)

    # Bin with default dr=0.1 Å
    centers, avg_mag = WannierHam.bin_hoppings_by_distance(entries, dr=0.1)

    # Should produce bins
    assert len(centers) > 0
    assert len(avg_mag) > 0
    assert len(centers) == len(avg_mag)

    # Bin centers should be monotonically increasing
    assert np.all(np.diff(centers) > 0), "Bin centers should be increasing"

    # Average magnitudes should be non-negative
    assert np.all(avg_mag >= 0), "Average magnitudes should be non-negative"

    # Try different bin width
    centers2, avg_mag2 = WannierHam.bin_hoppings_by_distance(entries, dr=0.5)
    assert len(centers2) < len(centers), "Larger dr should give fewer bins"


def test_bin_hoppings_empty():
    """Test binning with empty input."""
    centers, avg_mag = WannierHam.bin_hoppings_by_distance([], dr=0.1)

    assert len(centers) == 0
    assert len(avg_mag) == 0


def test_distance_mapping_consistency(srmo3_ham):
    """Test that Hamiltonian indices map consistently to distances."""
    ham, atoms = srmo3_ham
    cell = atoms.get_cell()

    entries = ham.distance_resolved_hoppings(cell)

    # Group by (R, i, j) to check uniqueness
    seen = set()
    for entry in entries:
        key = (entry["R"], entry["i"], entry["j"])
        assert key not in seen, (
            f"Duplicate hopping found: R={entry['R']}, i={entry['i']}, j={entry['j']}"
        )
        seen.add(key)

    # Verify distances are computed from Wannier positions
    positions_frac = np.asarray(ham.positions)

    # Check a few entries manually
    for entry in entries[:10]:  # Check first 10 for speed
        R = np.array(entry["R"])
        i = entry["i"]
        j = entry["j"]

        # Compute distance manually
        d_frac = positions_frac[j] + R - positions_frac[i]
        d_cart = d_frac @ cell
        expected_dist = float(np.linalg.norm(d_cart))

        assert np.isclose(entry["distance"], expected_dist, rtol=1e-6), (
            f"Distance mismatch for R={R}, i={i}, j={j}"
        )


def test_wannier_centers_exist(srmo3_ham):
    """Test that Wannier centers are available."""
    ham, _ = srmo3_ham

    # Should have positions attribute
    assert hasattr(ham, "positions")
    assert ham.positions is not None
    assert len(ham.positions) == ham.nbasis

    # Positions should be fractional coordinates in [0, 1)
    positions = np.asarray(ham.positions)
    assert positions.shape == (ham.nbasis, 3)



import math
import re
from collections import defaultdict

import numpy as np
from ase.atoms import Atoms
from ase.io import read
from ase.units import Angstrom, Bohr

from .utils import split_symbol_number
from .w90_tb_parser import parse_tb_file

unit_dict = {"ANG": Angstrom, "BOHR": Bohr}


def parse_xyz(fname):
    """
    wannier90 xyz file parser.

    :param fname: relative or absolute path to the xyz file.  str.
    """
    atoms = read(fname)
    symbols = atoms.get_chemical_symbols()
    pos = atoms.get_positions()
    wann_pos = []
    atoms_pos = []
    atoms_symbols = []
    for s, x in zip(symbols, pos):
        if s == "X":
            wann_pos.append(x)
        else:
            atoms_symbols.append(s)
            atoms_pos.append(x)
    return np.array(wann_pos), atoms_symbols, np.array(atoms_pos)


def parse_ham(fname="wannier90_hr.dat", cutoff=None):
    """
    wannier90 hr file parser.

    :param fname: relative or absolute path to the hamiltonian file.  str.

    :param cutoff: the energy cutoff.  None | number | list (of Emin, Emax).
    """
    with open(fname, "r") as myfile:
        lines = myfile.readlines()
    n_wann = int(lines[1].strip())
    n_R = int(lines[2].strip())

    # The lines of degeneracy of each R point. 15 per line.
    nline = int(math.ceil(n_R / 15.0))
    dlist = []
    for i in range(3, 3 + nline):
        d = map(float, lines[i].strip().split())
        dlist += d
    R_degens = np.array(dlist, dtype=int)
    H_mnR = defaultdict(lambda: np.zeros((n_wann, n_wann), dtype=complex))
    for i in range(3 + nline, 3 + nline + n_wann**2 * n_R):
        t = lines[i].strip().split()
        R = tuple(map(int, t[:3]))
        m, n = map(int, t[3:5])
        m = m - 1
        n = n - 1
        H_real, H_imag = map(float, t[5:])
        val = H_real + 1j * H_imag
        if m == n and np.linalg.norm(R) < 0.001:
            H_mnR[R][m, n] = val / 2.0
        elif cutoff is not None:
            if abs(val) > cutoff:
                H_mnR[R][m, n] = val / 2.0
        else:
            H_mnR[R][m, n] = val / 2.0
    return n_wann, H_mnR, R_degens


def parse_cell(fname, unit=Angstrom):
    """
    wannier90 hr cell parser.

    :param fname: relative or absolute path to the file.  str.
    """

    uc_regex = re.compile(
        r"BEGIN\s+UNIT_CELL_CART\s+"
        r"(?P<units>BOHR|ANG)?"
        r"(?P<cell>.+)"
        r"END\s+UNIT_CELL_CART\s+",
        re.VERBOSE | re.IGNORECASE | re.DOTALL,
    )
    with open(fname) as myfile:
        text = myfile.read()
    match = uc_regex.search(text)
    if match is None:
        raise Exception(f"Cannot find unit cell information from {fname}")

    cell = np.fromstring(match.group("cell").strip(), sep="\n").reshape((3, 3))

    if match.group("units") is not None:
        factor = {"ANG": Angstrom, "BOHR": Bohr}[match.group("units").upper()]
    else:
        factor = Angstrom
    cell = cell * factor / unit
    return cell


def parse_atoms(fname):
    """
    wannier90 hr atoms parser.

    :param fname: relative or absolute path to file.  str.
    """
    cell = parse_cell(fname)
    atoms_regex = re.compile(
        r"BEGIN\s+ATOMS_(?P<suffix>(FRAC)|(CART))\s+"
        r"(?P<units>BOHR|ANG)?"
        r"(?P<atoms>.+)"
        r"END\s+ATOMS_(?P=suffix)\s+",
        re.VERBOSE | re.IGNORECASE | re.DOTALL,
    )

    with open(fname, "r") as f:
        match = atoms_regex.search(f.read())
        if match is None:
            raise Exception(f"Cannot read atomic structure from {fname}")

    symbols = []
    taus = []
    tags = []

    for line in match.group("atoms").strip().splitlines():
        symbol = line.split()[0].strip()
        symbol = symbol[0].upper() + symbol[1:]
        symbol, tag = split_symbol_number(symbol)
        symbols.append(symbol)
        tags.append(tag)
        taus.append(np.array(list(map(float, line.split()[1:]))))

    taus = np.asarray(taus)
    if match.group("suffix").upper() == "FRAC":
        atoms = Atoms(symbols=symbols, cell=cell, scaled_positions=taus)
        atoms.set_tags(tags)
    else:
        if match.group("units") is not None:
            factor = unit_dict[match.group("units").upper()]
        else:
            factor = Angstrom
        taus = taus * factor
        atoms = Atoms(symbols=symbols, cell=cell, positions=taus)
    return atoms


def parse_tb(fname):
    """
    return the wannier center, number of wannier functions and the Hamiltonian from the wannier90 _tb.dat file.
    """
    result = parse_tb_file(fname)
    return result["centers"], result["n_wann"], result["H"], result["Rdegens"]



"""Plotting utilities for Wannier90 data."""

import os

from ase.io import read

from .wannier_hamiltonian import WannierHam


def plot_wannier_distance(
    ax,
    path,
    prefix,
    structure_file=None,
    structure_format=None,
    ylim=(1e-5, 10),
    **scatter_kwargs,
):
    """
    Plot Wannier hopping matrix elements vs. distance.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The matplotlib axes object to plot on.
    path : str
        Directory containing the Wannier90 files.
    prefix : str
        Prefix for Wannier90 files (e.g., "wannier90" for wannier90_hr.dat).
    structure_file : str, optional
        Structure file to read. If None, looks for common files.
    structure_format : str, optional
        Format of structure file (e.g., 'vasp', 'cif', 'espresso-in').
        If None, will be inferred from file extension.
    ylim : tuple, optional
        The y-axis limits for the plot. Default is (1e-5, 10).
    **scatter_kwargs : dict
        Additional keyword arguments to pass to `ax.scatter`.

    Returns
    -------
    ax : matplotlib.axes.Axes
        The configured matplotlib axes.
    """
    # Load structure
    if structure_file is None:
        # Try common structure file names
        for fname in ["POSCAR", "structure.cif", "input.pwi", "structure.vasp"]:
            fpath = os.path.join(path, fname)
            if os.path.exists(fpath):
                structure_file = fname
                break
        if structure_file is None:
            raise FileNotFoundError(
                "No structure file found. Please specify structure_file."
            )

    atoms = read(os.path.join(path, structure_file), format=structure_format)

    # Load Wannier Hamiltonian
    ham = WannierHam.read_from_wannier_dir(path, prefix, atoms=atoms)

    # Get distance-resolved hoppings
    from ase import Atoms

    if isinstance(atoms, Atoms):
        cell = atoms.get_cell()
    else:
        cell = atoms[0].get_cell()
    entries = ham.distance_resolved_hoppings(cell)

    # Extract data
    distances = [e["distance"] for e in entries]
    magnitudes = [abs(e["hopping"]) for e in entries]

    # Default scatter settings
    defaults = {"s": 5, "alpha": 0.6}
    defaults.update(scatter_kwargs)

    # Plot
    ax.scatter(distances, magnitudes, **defaults)
    ax.set_xlabel("Distance (Å)")
    ax.set_ylabel("|t| (eV)")
    ax.set_yscale("log")
    ax.set_ylim(ylim)

    return ax



"""Tests for distance-resolved coupling analysis in EPW electron-phonon data.

Tests cover:
    - Distance computation from Wannier centers (Rk-based)
    - Distance computation from atomic positions (Rg-based)
    - Binning of coupling magnitudes by distance
    - Correctness of mode-to-atom mapping
"""

from pathlib import Path

import numpy as np
import pytest
from ase.units import Bohr

from HamiltonIO.epw.epwparser import Epmat, read_crystal_fmt


@pytest.fixture
def srmo3_epmat():
    """Load SrMnO3 EPW data from test data."""
    # Find test data directory
    project_root = Path(__file__).parent.parent
    data_dir = project_root / "HamiltonIO" / "epw" / "test" / "up"

    # Read crystal structure
    crystal = read_crystal_fmt(str(data_dir / "crystal.fmt"))

    # Read EPW data
    ep = Epmat()
    ep.crystal = crystal
    ep.read(path=str(data_dir), prefix="SrMnO3", epmat_ncfile="epmat.nc")

    # Get cell in Angstrom
    cell = crystal.at.reshape(3, 3) * crystal.alat * Bohr

    return ep, crystal, cell


def test_epmat_loading(srmo3_epmat):
    """Test that EPW data loads correctly."""
    ep, crystal, cell = srmo3_epmat

    # Check basic properties
    assert ep.nwann > 0, "Should have Wannier functions"
    assert crystal.natom > 0, "Should have atoms"
    assert crystal.nmode == crystal.natom * 3, "Should have 3 modes per atom"

    # Check that data is loaded
    assert hasattr(ep, "epmatfile") or hasattr(ep, "epmat_wann"), (
        "Should have matrix data"
    )
    assert ep.nRk > 0, "Should have k-space R vectors"
    assert ep.nRg > 0, "Should have g-space R vectors"

    # Check cell shape
    assert cell.shape == (3, 3), "Cell should be 3x3 matrix"


def test_distance_resolved_couplings_Rk(srmo3_epmat):
    """Test Rk-based distance computation (WF to WF)."""
    ep, crystal, cell = srmo3_epmat
    imode = 0

    # Get distance-resolved couplings
    entries = ep.distance_resolved_couplings_Rk(imode=imode, cell=cell)

    # Should have couplings
    assert len(entries) > 0, "Should have non-zero coupling elements"

    # Check structure of each entry
    for entry in entries:
        assert "imode" in entry
        assert "Rg" in entry
        assert "Rk" in entry
        assert "i" in entry
        assert "j" in entry
        assert "distance" in entry
        assert "coupling" in entry

        # Check types
        assert entry["imode"] == imode
        assert len(entry["Rg"]) == 3
        assert len(entry["Rk"]) == 3
        assert isinstance(entry["i"], int)
        assert isinstance(entry["j"], int)
        assert isinstance(entry["distance"], float)
        assert entry["distance"] >= 0, "Distance should be non-negative"
        assert isinstance(entry["coupling"], complex)

    # Distances should be reasonable
    distances = np.array([e["distance"] for e in entries])
    assert distances.min() >= 0
    assert distances.max() < 100, "Distances should be reasonable (< 100 Å)"
    assert len(np.unique(distances)) > 1, "Should have multiple distinct distances"


def test_distance_resolved_couplings_Rg(srmo3_epmat):
    """Test Rg-based distance computation (WF to atom)."""
    ep, crystal, cell = srmo3_epmat
    imode = 0

    # Get distance-resolved couplings
    entries = ep.distance_resolved_couplings_Rg(imode=imode, cell=cell)

    # Should have couplings
    assert len(entries) > 0, "Should have non-zero coupling elements"

    # Check structure of each entry
    for entry in entries:
        assert "imode" in entry
        assert "atom_index" in entry
        assert "Rg" in entry
        assert "Rk" in entry
        assert "i" in entry
        assert "j" in entry
        assert "distance" in entry
        assert "coupling" in entry

        # Check types
        assert entry["imode"] == imode
        assert isinstance(entry["atom_index"], int)
        assert 0 <= entry["atom_index"] < crystal.natom, "Atom index should be valid"
        assert len(entry["Rg"]) == 3
        assert len(entry["Rk"]) == 3
        assert isinstance(entry["i"], int)
        assert isinstance(entry["j"], int)
        assert isinstance(entry["distance"], float)
        assert entry["distance"] >= 0, "Distance should be non-negative"
        assert isinstance(entry["coupling"], complex)

    # Distances should be reasonable
    distances = np.array([e["distance"] for e in entries])
    assert distances.min() >= 0
    assert distances.max() < 100, "Distances should be reasonable (< 100 Å)"


def test_mode_to_atom_mapping(srmo3_epmat):
    """Test that mode-to-atom mapping is correct."""
    ep, crystal, cell = srmo3_epmat

    # Test several modes
    for imode in range(
        min(9, crystal.nmode)
    ):  # Test first 9 modes (3 atoms × 3 directions)
        entries = ep.distance_resolved_couplings_Rg(imode=imode, cell=cell)

        # All entries for this mode should map to the same atom
        atom_indices = set(e["atom_index"] for e in entries)
        assert len(atom_indices) == 1, f"Mode {imode} should map to single atom"

        # Verify the mapping: atom_index = imode // 3
        expected_atom = imode // 3
        actual_atom = atom_indices.pop()
        assert actual_atom == expected_atom, (
            f"Mode {imode} should map to atom {expected_atom}, got {actual_atom}"
        )


def test_bin_couplings_by_distance(srmo3_epmat):
    """Test binning of coupling magnitudes by distance."""
    ep, crystal, cell = srmo3_epmat
    imode = 0

    entries = ep.distance_resolved_couplings_Rk(imode=imode, cell=cell)

    # Bin with default dr=0.1 Å
    centers, avg_mag = Epmat.bin_couplings_by_distance(entries, dr=0.1)

    # Should produce bins
    assert len(centers) > 0, "Should have bin centers"
    assert len(avg_mag) > 0, "Should have average magnitudes"
    assert len(centers) == len(avg_mag), "Should have same number of centers and values"

    # Bin centers should be monotonically increasing
    assert np.all(np.diff(centers) > 0), "Bin centers should be increasing"

    # Average magnitudes should be non-negative
    assert np.all(avg_mag >= 0), "Average magnitudes should be non-negative"

    # Try different bin width
    centers2, avg_mag2 = Epmat.bin_couplings_by_distance(entries, dr=0.5)
    assert len(centers2) < len(centers), "Larger dr should give fewer bins"


def test_bin_couplings_empty():
    """Test binning with empty input."""
    centers, avg_mag = Epmat.bin_couplings_by_distance([], dr=0.1)

    assert len(centers) == 0, "Empty input should give no bins"
    assert len(avg_mag) == 0, "Empty input should give no values"


def test_Rk_vs_Rg_consistency(srmo3_epmat):
    """Test that Rk and Rg methods give same coupling values but different distances."""
    ep, crystal, cell = srmo3_epmat
    imode = 0

    entries_Rk = ep.distance_resolved_couplings_Rk(imode=imode, cell=cell)
    entries_Rg = ep.distance_resolved_couplings_Rg(imode=imode, cell=cell)

    # Should have same number of entries
    assert len(entries_Rk) == len(entries_Rg), (
        "Rk and Rg methods should return same number of entries"
    )

    # Create lookup dictionaries
    Rk_lookup = {(e["Rg"], e["Rk"], e["i"], e["j"]): e for e in entries_Rk}
    Rg_lookup = {(e["Rg"], e["Rk"], e["i"], e["j"]): e for e in entries_Rg}

    # Check that coupling values match
    matches = 0
    dist_diffs = 0
    for key in Rk_lookup:
        if key in Rg_lookup:
            matches += 1
            # Coupling values should be identical
            assert np.isclose(
                Rk_lookup[key]["coupling"], Rg_lookup[key]["coupling"], rtol=1e-10
            ), f"Coupling mismatch for {key}"

            # Distances should generally be different (unless special case)
            if Rk_lookup[key]["distance"] != Rg_lookup[key]["distance"]:
                dist_diffs += 1

    assert matches > 0, "Should have matching entries"
    # Most distances should differ between Rk and Rg
    assert dist_diffs > matches * 0.5, "Most Rk and Rg distances should differ"


def test_onsite_Rk_distances(srmo3_epmat):
    """Test that Rk=0, i=j couplings have near-zero distance."""
    ep, crystal, cell = srmo3_epmat
    imode = 0

    entries = ep.distance_resolved_couplings_Rk(imode=imode, cell=cell)

    # Find onsite couplings (Rk=0, i=j)
    onsite = [e for e in entries if e["Rk"] == (0, 0, 0) and e["i"] == e["j"]]

    # Should have some onsite terms
    assert len(onsite) > 0, "Should have onsite diagonal elements"

    # Onsite distances should be very small (essentially zero)
    for entry in onsite:
        assert entry["distance"] < 1e-6, (
            f"Onsite Rk distance should be ~0, got {entry['distance']}"
        )


def test_wannier_centers_available(srmo3_epmat):
    """Test that Wannier centers are available in crystal structure."""
    ep, crystal, cell = srmo3_epmat

    # Should have Wannier centers
    assert hasattr(crystal, "w_centers"), "Crystal should have w_centers"
    assert crystal.w_centers is not None, "w_centers should not be None"

    # Check shape
    w_centers = np.asarray(crystal.w_centers).reshape(ep.nwann, 3)
    assert w_centers.shape == (ep.nwann, 3), (
        f"w_centers should be ({ep.nwann}, 3), got {w_centers.shape}"
    )


def test_atomic_positions_available(srmo3_epmat):
    """Test that atomic positions are available in crystal structure."""
    ep, crystal, cell = srmo3_epmat

    # Should have atomic positions
    assert hasattr(crystal, "tau"), "Crystal should have tau (atomic positions)"
    assert crystal.tau is not None, "tau should not be None"

    # Check shape
    tau = np.asarray(crystal.tau).reshape(crystal.natom, 3)
    assert tau.shape == (crystal.natom, 3), (
        f"tau should be ({crystal.natom}, 3), got {tau.shape}"
    )



import numpy as np

_all__ = ["HR_to_k", "Hk_to_R", "k_to_R", "R_to_k", "R_to_onek"]


def HR_to_k(HR, Rlist, kpts):
    # Hk[k,:,:] = sum_R (H[R] exp(i2pi k.R))
    phase = np.exp(2.0j * np.pi * np.tensordot(kpts, Rlist, axes=([1], [1])))
    Hk = np.einsum("rlm, kr -> klm", HR, phase)
    return Hk


def Hk_to_R(Hk, Rlist, kpts, kweights):
    phase = np.exp(-2.0j * np.pi * np.tensordot(kpts, Rlist, axes=([1], [1])))
    HR = np.einsum("klm, kr, k->rlm", Hk, phase, kweights)
    return HR


def k_to_R(kpts, Rlist, Mk, kweights=None):
    """
    Transform k-space wavefunctions to real space.
    params:
        kpts: k-points
        Rlist: list of R vectors
        Mk: matrix of shape [nkpt, n1, n2] in k-space.

    return:
        MR: matrix of shape [nR, n1, n2], the matrix in R-space.

    """
    nkpt, n1, n2 = Mk.shape
    if kweights is None:
        kweights = np.ones(nkpt, dtype=float) / nkpt
    phase = np.exp(-2.0j * np.pi * np.tensordot(kpts, Rlist, axes=([1], [1])))
    MR = np.einsum("klm, kr, k -> rlm", Mk, phase, kweights)
    return MR

    # nkpt, n1, n2 = Mk.shape
    # nR = Rlist.shape[0]
    # MR = np.zeros((nR, n1, n2), dtype=complex)
    # if kweights is None:
    #    kweights = np.ones(nkpt, dtype=float)/nkpt
    # for iR, R in enumerate(Rlist):
    #    for ik in range(nkpt):
    #        MR[iR] += Mk[ik] * np.exp(-2.0j*np.pi * np.dot(kpts[ik], R)) * kweights[ik]
    # return MR


def R_to_k(kpts, Rlist, MR):
    """
    Transform real-space wavefunctions to k-space.
    params:
        kpts: k-points
        Rlist: list of R vectors
        MR: matrix of shape [nR, n1, n2] in R-space.

    return:
        Mk: matrix of shape [nkpt, n1, n2], the matrix in k-space.

    """
    phase = np.exp(2.0 * np.pi * 1j * np.tensordot(kpts, Rlist, axes=([1], [1])))
    Mk = np.einsum("rlm, kr -> klm", MR, phase)

    # nkpt, n1, n2 = Mk.shape
    # nR = Rlist.shape[0]
    # Mk = np.zeros((nkpt, n1, n2), dtype=complex)
    # for iR, R in enumerate(Rlist):
    #    for ik in range(nkpt):
    #        Mk[ik] += MR[iR] * np.exp(2.0 * np.pi * 1j * np.dot(kpts[ik], R))
    return Mk


def R_to_onek(kpt, Rlist, MR):
    """
    Transform real-space wavefunctions to k-space.
    params:
        kpt: k-point
        Rlist: list of R vectors
        MR: matrix of shape [nR, n1, n2] in R-space.

    return:
        Mk: matrix of shape [n1, n2], the matrix in k-space.

    """
    phase = np.exp(2.0j * np.pi * np.dot(Rlist, kpt))
    Mk = np.einsum("rlm, r -> lm", MR, phase)
    return Mk
    # n1, n2 = MR.shape[1:]
    #    Mk = np.zeros((n1, n2), dtype=complex)
    #    for iR, R in enumerate(Rlist):
    #        Mk += MR[iR] * np.exp(2.0j*np.pi * np.dot(kpt, R))
    #    return Mk



import sisl
import numpy as np
import matplotlib.pyplot as plt
import os
import sys

############################################################################
def calctype(data):
    
    if data == 'unpolarized':
     outdata = 'NP'
        
    elif data == 'polarized':
     outdata = 'SP'
        
    elif data == 'non-colinear':
     outdata = 'SOC'
        
    elif data == 'spin-orbit':
     outdata = 'SOC'
            
    return outdata   


def ncases(var):
    return "%8f"%var

###############################################################################

inputfdf=  'MoS2_SZP_SOC/RUN.fdf' 

#inputfdf=  sys.argv[1]  

tshs = sisl.get_sile(inputfdf).read_hamiltonian() #pegar hamiltoniano do siesta

fermi = sisl.get_sile(inputfdf).read_fermi_level()

scs = 0.0000 #scissors operator

nbasis= tshs.no
ncell=  tshs.nsc[0]*tshs.nsc[1]*tshs.nsc[2]
sptype=str(tshs.spin)

sptype2=sptype[5:-9]

if sptype2 == 'unpolarized' :

 f = open("system-info-NP.txt", "a")
 print(tshs, file=f)
 f.close()

 nbasis= tshs.no
 ncell=  tshs.nsc[0]*tshs.nsc[1]*tshs.nsc[2]
 sptype=str(tshs.spin)


 f = open("siesta_tb-NP.dat", "a")

 print(calctype(sptype2),file=f)
 print(ncases(scs),file=f)
 print(ncases(fermi[0]),file=f)
 print(ncases(tshs.cell[0,0]),'',ncases(tshs.cell[0,1]),'',ncases(tshs.cell[0,2]),file=f)
 print(ncases(tshs.cell[1,0]),'',ncases(tshs.cell[1,1]),'',ncases(tshs.cell[1,2]),file=f)
 print(ncases(tshs.cell[2,0]),'',ncases(tshs.cell[2,1]),'',ncases(tshs.cell[2,2]),file=f)
 print(nbasis,file=f)
 print(ncell,file=f)
 print('#rcell x',' ','rcell y',' ','rcell z',' ','i',' ','j',' ','ReH',' ','ImH',' ','S',file=f)

 for i in range(0,ncell):
  for j in range(0,nbasis): 
   for k in range(0,nbasis):
  	
    a = tshs.geometry.o2sc(k+i*nbasis)[0]
    b = tshs.geometry.o2sc(k+i*nbasis)[1]
    c = tshs.geometry.o2sc(k+i*nbasis)[2]
		
    d = tshs.H[j,k+i*nbasis]
    e = tshs.S[j,k+i*nbasis]
	
    print(ncases(a),' ',ncases(b), ' ',ncases(c), ' ',j+1,' ',k+1,' ',ncases(d),' ',ncases(0.00),' ',ncases(e),file=f)


 f.close()

 f = open("siesta_basis_set-NP", "a")

 print('bindex aspecie ax ay az l m spin',file=f)

 io = 0
 for ia, atom in enumerate(tshs.geometry.atoms):
    xyz = tshs.geometry.xyz[ia, :]
    for orbital in atom:
        print(io+1,' ', atom.tag,' ', ncases(xyz[0]),' ', ncases(xyz[1]),' ', ncases(xyz[2]),' ', orbital.l,' ', orbital.m,' ',0,file=f)
        io += 1

 f.close()


####################################################################

if sptype2 == 'polarized' :

 f = open("system-info-sp.txt", "a")
 print(tshs, file=f)
 f.close()

 nbasis= tshs.no
 ncell=  tshs.nsc[0]*tshs.nsc[1]*tshs.nsc[2]
 sptype=str(tshs.spin)


 f = open("siesta_basis_set-sp", "a")

 print('#bindex aspecie ax ay az l m spin',file=f)

 io = 0
 for ia, atom in enumerate(tshs.geometry.atoms):
    xyz = tshs.geometry.xyz[ia, :]
    for orbital in atom:
        print(io+1,' ', atom.tag,' ', ncases(xyz[0]),' ', ncases(xyz[1]),' ', ncases(xyz[2]),' ', orbital.l,' ', orbital.m,' ',1,file=f)
        io += 1

 io = 0
 for ia, atom in enumerate(tshs.geometry.atoms):
    xyz = tshs.geometry.xyz[ia, :]
    for orbital in atom:
        print(nbasis+io+1,' ', atom.tag,' ', ncases(xyz[0]),' ', ncases(xyz[1]),' ', ncases(xyz[2]),' ', orbital.l,' ', orbital.m,' ',-1,file=f)
        io += 1

 f.close()


 f = open("siesta_tb-sp.dat", "a")

 print(calctype(sptype2),file=f)
 print(ncases(scs),file=f)
 print(ncases(fermi[0]),file=f)
 print(ncases(tshs.cell[0,0]),'',ncases(tshs.cell[0,1]),'',ncases(tshs.cell[0,2]),file=f)
 print(ncases(tshs.cell[1,0]),'',ncases(tshs.cell[1,1]),'',ncases(tshs.cell[1,2]),file=f)
 print(ncases(tshs.cell[2,0]),'',ncases(tshs.cell[2,1]),'',ncases(tshs.cell[2,2]),file=f)
 print(2*nbasis,file=f)
 print(ncell,file=f)
 print('#rcell x',' ','rcell y',' ','rcell z',' ','i',' ','j',' ','ReH',' ','ImH',' ','S',file=f)
#np.empty aloca todos os arrays vazios
#alocando todos os arrays com 0

 S = np.zeros((ncell+1,(2*nbasis)+2,(2*nbasis)+2))
 reH = np.zeros((ncell+1,(2*nbasis)+2,(2*nbasis)+2))
 imH = np.zeros((ncell+1,(2*nbasis)+2,(2*nbasis)+2))

 a = np.zeros((ncell+1,(2*nbasis)+2))
 b = np.zeros((ncell+1,(2*nbasis)+2))
 c = np.zeros((ncell+1,(2*nbasis)+2))


 for i in range(0,ncell):
  for j in range(0,nbasis): 
   for k in range(0,nbasis):

#parte up-up
  
    reH[i+1,j+1,k+1] = tshs[j,k+i*nbasis][0]
    #imH[i+1,j+1,k+1] = tshs[j,k+i*nbasis][4]
    S[i+1,j+1,k+1] = tshs.S[j,k+i*nbasis]

    

#parte up-dn

    #reH[i+1,(j+1),(k+1)+nbasis] = tshs[j,k+i*nbasis][2]
    #imH[i+1,(j+1),(k+1)+nbasis] = tshs[j,k+i*nbasis][3]
    #S[i+1,(j+1),(k+1)+nbasis] = 0.0

#parte dn-up

    #reH[i+1,(j+1)+nbasis,(k+1)] = tshs[j,k+i*nbasis][6]
    #imH[i+1,(j+1)+nbasis,(k+1)] = tshs[j,k+i*nbasis][7]
    #S[i+1,(j+1)+nbasis,(k+1)] = 0.0

#parte dn-dn

    reH[i+1,(j+1)+nbasis,(k+1)+nbasis] = tshs[j,k+i*nbasis][1]
    #imH[i+1,(j+1)+nbasis,(k+1)+nbasis] = tshs[j,k+i*nbasis][5]
    S[i+1,(j+1)+nbasis,(k+1)+nbasis] = tshs.S[j,k+i*nbasis]
    
#escrevendo a localização dos atomos da base

 for i in range(0,ncell):
  for k in range(0,nbasis): 

   a[i+1,k+1] = tshs.geometry.o2sc(k+i*nbasis)[0]
   b[i+1,k+1] = tshs.geometry.o2sc(k+i*nbasis)[1]
   c[i+1,k+1] = tshs.geometry.o2sc(k+i*nbasis)[2]
	
   a[i+1,(k+1)+nbasis] = tshs.geometry.o2sc(k+i*nbasis)[0]
   b[i+1,(k+1)+nbasis] = tshs.geometry.o2sc(k+i*nbasis)[1]
   c[i+1,(k+1)+nbasis] = tshs.geometry.o2sc(k+i*nbasis)[2]
	

 for i in range(1,ncell+1):
  for j in range(1,(2*nbasis)+1): 
   for k in range(1,(2*nbasis)+1):
  	
#   a = tshs.geometry.o2sc(k+i*nbasis)[0]
#   b = tshs.geometry.o2sc(k+i*nbasis)[1]
#   c = tshs.geometry.o2sc(k+i*nbasis)[2]
		
#   d = tshs.H[j,k+i*nbasis]
#   e = tshs.S[j,k+i*nbasis]
	
    print(ncases(a[i,k]),' ',ncases(b[i,k]), ' ',ncases(c[i,k]), ' ',j,' ',k,' ',ncases(reH[i,j,k]),' ',ncases(imH[i,j,k]),' ',ncases(S[i,j,k]),file=f)


 f.close()

####################################################################

if sptype2 == 'non-colinear' :

 f = open("system-info-nc.txt", "a")
 print(tshs, file=f)
 f.close()

 nbasis= tshs.no
 ncell=  tshs.nsc[0]*tshs.nsc[1]*tshs.nsc[2]
 sptype=str(tshs.spin)


 f = open("siesta_basis_set-nc", "a")

 print('#bindex aspecie ax ay az l m spin',file=f)

 io = 0
 for ia, atom in enumerate(tshs.geometry.atoms):
    xyz = tshs.geometry.xyz[ia, :]
    for orbital in atom:
        print(io+1,' ', atom.tag,' ', ncases(xyz[0]),' ', ncases(xyz[1]),' ', ncases(xyz[2]),' ', orbital.l,' ', orbital.m,' ',1,file=f)
        io += 1

 io = 0
 for ia, atom in enumerate(tshs.geometry.atoms):
    xyz = tshs.geometry.xyz[ia, :]
    for orbital in atom:
        print(nbasis+io+1,' ', atom.tag,' ', ncases(xyz[0]),' ', ncases(xyz[1]),' ', ncases(xyz[2]),' ', orbital.l,' ', orbital.m,' ',-1,file=f)
        io += 1

 f.close()


 f = open("siesta_tb-nc.dat", "a")

 print(calctype(sptype2),file=f)
 print(ncases(scs),file=f)
 print(ncases(fermi[0]),file=f)
 print(ncases(tshs.cell[0,0]),'',ncases(tshs.cell[0,1]),'',ncases(tshs.cell[0,2]),file=f)
 print(ncases(tshs.cell[1,0]),'',ncases(tshs.cell[1,1]),'',ncases(tshs.cell[1,2]),file=f)
 print(ncases(tshs.cell[2,0]),'',ncases(tshs.cell[2,1]),'',ncases(tshs.cell[2,2]),file=f)
 print(2*nbasis,file=f)
 print(ncell,file=f)
 print('#rcell x',' ','rcell y',' ','rcell z',' ','i',' ','j',' ','ReH',' ','ImH',' ','S',file=f)

#np.empty aloca todos os arrays vazios
#alocando todos os arrays com 0

 S = np.zeros((ncell+1,(2*nbasis)+2,(2*nbasis)+2))
 reH = np.zeros((ncell+1,(2*nbasis)+2,(2*nbasis)+2))
 imH = np.zeros((ncell+1,(2*nbasis)+2,(2*nbasis)+2))

 a = np.zeros((ncell+1,(2*nbasis)+2))
 b = np.zeros((ncell+1,(2*nbasis)+2))
 c = np.zeros((ncell+1,(2*nbasis)+2))


 for i in range(0,ncell):
  for j in range(0,nbasis): 
   for k in range(0,nbasis):

#parte up-up
  
     reH[i+1,j+1,k+1] = tshs[j,k+i*nbasis][0]
    #imH[i+1,j+1,k+1] = tshs[j,k+i*nbasis][4]
     S[i+1,j+1,k+1] = tshs.S[j,k+i*nbasis]

    

#parte up-dn

     reH[i+1,(j+1),(k+1)+nbasis] = tshs[j,k+i*nbasis][2]
    #imH[i+1,(j+1),(k+1)+nbasis] = tshs[j,k+i*nbasis][3]
     S[i+1,(j+1),(k+1)+nbasis] = 0.0

#parte dn-up

     reH[i+1,(j+1)+nbasis,(k+1)] = tshs[j,k+i*nbasis][3]
    #imH[i+1,(j+1)+nbasis,(k+1)] = tshs[j,k+i*nbasis][7]
     S[i+1,(j+1)+nbasis,(k+1)] = 0.0

#parte dn-dn

     reH[i+1,(j+1)+nbasis,(k+1)+nbasis] = tshs[j,k+i*nbasis][1]
    #imH[i+1,(j+1)+nbasis,(k+1)+nbasis] = tshs[j,k+i*nbasis][5]
     S[i+1,(j+1)+nbasis,(k+1)+nbasis] = tshs.S[j,k+i*nbasis]
    
#escrevendo a localização dos atomos da base

 for i in range(0,ncell):
  for k in range(0,nbasis): 

   a[i+1,k+1] = tshs.geometry.o2sc(k+i*nbasis)[0]
   b[i+1,k+1] = tshs.geometry.o2sc(k+i*nbasis)[1]
   c[i+1,k+1] = tshs.geometry.o2sc(k+i*nbasis)[2]
	
   a[i+1,(k+1)+nbasis] = tshs.geometry.o2sc(k+i*nbasis)[0]
   b[i+1,(k+1)+nbasis] = tshs.geometry.o2sc(k+i*nbasis)[1]
   c[i+1,(k+1)+nbasis] = tshs.geometry.o2sc(k+i*nbasis)[2]
	

 for i in range(1,ncell+1):
  for j in range(1,(2*nbasis)+1): 
   for k in range(1,(2*nbasis)+1):
  	
#   a = tshs.geometry.o2sc(k+i*nbasis)[0]
#   b = tshs.geometry.o2sc(k+i*nbasis)[1]
#   c = tshs.geometry.o2sc(k+i*nbasis)[2]
		
#   d = tshs.H[j,k+i*nbasis]
#   e = tshs.S[j,k+i*nbasis]
	
    print(ncases(a[i,k]),' ',ncases(b[i,k]), ' ',ncases(c[i,k]), ' ',j,' ',k,' ',ncases(reH[i,j,k]),' ',ncases(imH[i,j,k]),' ',ncases(S[i,j,k]),file=f)


 f.close()

####################################################################

if sptype2 == 'spin-orbit' :

 f = open("siesta_basis_set-soc", "a")

 print('#bindex aspecie ax ay az l m spin',file=f)

 io = 0
 for ia, atom in enumerate(tshs.geometry.atoms):
     xyz = tshs.geometry.xyz[ia, :]
     for orbital in atom:
         print(io+1,' ', atom.tag,' ', ncases(xyz[0]),' ', ncases(xyz[1]),' ', ncases(xyz[2]),' ', orbital.l,' ', orbital.m,' ',1,file=f)
         io += 1

 io = 0
 for ia, atom in enumerate(tshs.geometry.atoms):
     xyz = tshs.geometry.xyz[ia, :]
     for orbital in atom:
         print(nbasis+io+1,' ', atom.tag,' ', ncases(xyz[0]),' ', ncases(xyz[1]),' ', ncases(xyz[2]),' ', orbital.l,' ', orbital.m,' ',-1,file=f)
         io += 1

 f.close()


 f = open("siesta_tb-soc.dat", "a")

 print(calctype(sptype2),file=f)
 print(ncases(scs),file=f)
 print(ncases(fermi[0]),file=f)
 print(ncases(tshs.cell[0,0]),'',ncases(tshs.cell[0,1]),'',ncases(tshs.cell[0,2]),file=f)
 print(ncases(tshs.cell[1,0]),'',ncases(tshs.cell[1,1]),'',ncases(tshs.cell[1,2]),file=f)
 print(ncases(tshs.cell[2,0]),'',ncases(tshs.cell[2,1]),'',ncases(tshs.cell[2,2]),file=f)
 print(2*nbasis,file=f)
 print(ncell,file=f)
 print('#rcell x',' ','rcell y',' ','rcell z',' ','i',' ','j',' ','ReH',' ','ImH',' ','S',file=f)

#np.empty aloca todos os arrays vazios
#alocando todos os arrays com 0

 S = np.zeros((ncell+1,(2*nbasis)+2,(2*nbasis)+2))
 reH = np.zeros((ncell+1,(2*nbasis)+2,(2*nbasis)+2))
 imH = np.zeros((ncell+1,(2*nbasis)+2,(2*nbasis)+2))

 a = np.zeros((ncell+1,(2*nbasis)+2))
 b = np.zeros((ncell+1,(2*nbasis)+2))
 c = np.zeros((ncell+1,(2*nbasis)+2))


 for i in range(0,ncell):
  for j in range(0,nbasis): 
   for k in range(0,nbasis):

#parte up-up
  
    reH[i+1,j+1,k+1] = tshs[j,k+i*nbasis][0]
    imH[i+1,j+1,k+1] = tshs[j,k+i*nbasis][4]
    S[i+1,j+1,k+1] = tshs.S[j,k+i*nbasis]

    

#parte up-dn

    reH[i+1,(j+1),(k+1)+nbasis] = tshs[j,k+i*nbasis][2]
    imH[i+1,(j+1),(k+1)+nbasis] = tshs[j,k+i*nbasis][3]
    S[i+1,(j+1),(k+1)+nbasis] = 0.0

#parte dn-up

    reH[i+1,(j+1)+nbasis,(k+1)] = tshs[j,k+i*nbasis][6]
    imH[i+1,(j+1)+nbasis,(k+1)] = tshs[j,k+i*nbasis][7]
    S[i+1,(j+1)+nbasis,(k+1)] = 0.0

#parte dn-dn

    reH[i+1,(j+1)+nbasis,(k+1)+nbasis] = tshs[j,k+i*nbasis][1]
    imH[i+1,(j+1)+nbasis,(k+1)+nbasis] = tshs[j,k+i*nbasis][5]
    S[i+1,(j+1)+nbasis,(k+1)+nbasis] = tshs.S[j,k+i*nbasis]
    
#escrevendo a localização dos atomos da base

 for i in range(0,ncell):
  for k in range(0,nbasis): 

   a[i+1,k+1] = tshs.geometry.o2sc(k+i*nbasis)[0]
   b[i+1,k+1] = tshs.geometry.o2sc(k+i*nbasis)[1]
   c[i+1,k+1] = tshs.geometry.o2sc(k+i*nbasis)[2]
	
   a[i+1,(k+1)+nbasis] = tshs.geometry.o2sc(k+i*nbasis)[0]
   b[i+1,(k+1)+nbasis] = tshs.geometry.o2sc(k+i*nbasis)[1]
   c[i+1,(k+1)+nbasis] = tshs.geometry.o2sc(k+i*nbasis)[2]
	

 for i in range(1,ncell+1):
  for j in range(1,(2*nbasis)+1): 
   for k in range(1,(2*nbasis)+1):
  	
#   a = tshs.geometry.o2sc(k+i*nbasis)[0]
#   b = tshs.geometry.o2sc(k+i*nbasis)[1]
#   c = tshs.geometry.o2sc(k+i*nbasis)[2]
		
#   d = tshs.H[j,k+i*nbasis]
#   e = tshs.S[j,k+i*nbasis]
	
    print(ncases(a[i,k]),' ',(b[i,k]), ' ',ncases(c[i,k]), ' ',j,' ',k,' ',ncases(reH[i,j,k]),' ',ncases(imH[i,j,k]),' ',ncases(S[i,j,k]),file=f)


 f.close()

	
####################################################################	




#!/usr/bin/env python3
"""
Script to remove '.down' suffix from filenames in a directory.

This script will rename files by removing '.down' from their names.
For example:
- 'crystal.down.fmt' -> 'crystal.fmt'
- 'epwdata.down.fmt' -> 'epwdata.fmt'
- 'wigner.down.fmt' -> 'wigner.fmt'

Usage:
    python remove_down_suffix.py [directory] [options]

Options:
    -d, --directory DIR    Directory to process (default: current directory)
    -n, --dry-run         Show what would be renamed without actually doing it
    -v, --verbose         Show detailed output
    -r, --recursive       Process subdirectories recursively
    -h, --help           Show this help message
"""

import os
import sys
import argparse
import glob
from pathlib import Path


def remove_down_from_filename(filepath, dry_run=False, verbose=False):
    """
    Remove '.down' from a filename.
    
    Parameters
    ----------
    filepath : str or Path
        Path to the file to rename
    dry_run : bool, optional
        If True, only show what would be done without actually renaming
    verbose : bool, optional
        If True, show detailed output
        
    Returns
    -------
    bool
        True if file was renamed (or would be renamed in dry-run), False otherwise
    """
    filepath = Path(filepath)
    
    # Check if filename contains '.down'
    if '.down' not in filepath.name:
        if verbose:
            print(f"   Skip: {filepath.name} (no '.down' found)")
        return False
    
    # Create new filename by removing '.down'
    new_name = filepath.name.replace('.down', '')
    new_filepath = filepath.parent / new_name
    
    # Check if target file already exists
    if new_filepath.exists():
        print(f"   ⚠️  Warning: Target file already exists: {new_filepath}")
        print(f"      Skipping: {filepath}")
        return False
    
    if dry_run:
        print(f"   Would rename: {filepath.name} -> {new_name}")
        return True
    else:
        try:
            filepath.rename(new_filepath)
            if verbose:
                print(f"   ✅ Renamed: {filepath.name} -> {new_name}")
            else:
                print(f"   {filepath.name} -> {new_name}")
            return True
        except Exception as e:
            print(f"   ❌ Error renaming {filepath}: {e}")
            return False


def process_directory(directory, dry_run=False, verbose=False, recursive=False):
    """
    Process all files in a directory to remove '.down' from filenames.
    
    Parameters
    ----------
    directory : str or Path
        Directory to process
    dry_run : bool, optional
        If True, only show what would be done
    verbose : bool, optional
        If True, show detailed output
    recursive : bool, optional
        If True, process subdirectories recursively
        
    Returns
    -------
    tuple
        (total_files_processed, files_renamed)
    """
    directory = Path(directory)
    
    if not directory.exists():
        print(f"❌ Directory does not exist: {directory}")
        return 0, 0
    
    if not directory.is_dir():
        print(f"❌ Not a directory: {directory}")
        return 0, 0
    
    print(f"Processing directory: {directory.absolute()}")
    if dry_run:
        print("🔍 DRY RUN MODE - No files will be actually renamed")
    print()
    
    total_files = 0
    renamed_files = 0
    
    # Get file pattern based on recursive option
    if recursive:
        pattern = "**/*"
        files = directory.glob(pattern)
        files = [f for f in files if f.is_file()]  # Filter only files
    else:
        files = [f for f in directory.iterdir() if f.is_file()]
    
    # Filter files that contain '.down'
    down_files = [f for f in files if '.down' in f.name]
    
    if not down_files:
        print("No files with '.down' found in the directory.")
        return 0, 0
    
    print(f"Found {len(down_files)} file(s) with '.down' in the name:")
    print()
    
    for filepath in sorted(down_files):
        total_files += 1
        if remove_down_from_filename(filepath, dry_run, verbose):
            renamed_files += 1
    
    return total_files, renamed_files


def main():
    """Main function to handle command line arguments and execute the script."""
    parser = argparse.ArgumentParser(
        description="Remove '.down' suffix from filenames in a directory",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
    # Process current directory
    python remove_down_suffix.py
    
    # Process specific directory
    python remove_down_suffix.py -d /path/to/directory
    
    # Dry run to see what would be renamed
    python remove_down_suffix.py -n
    
    # Process recursively with verbose output
    python remove_down_suffix.py -r -v
    
    # Process test/down directory
    python remove_down_suffix.py -d test/down
        """
    )
    
    parser.add_argument(
        '-d', '--directory',
        default='.',
        help='Directory to process (default: current directory)'
    )
    
    parser.add_argument(
        '-n', '--dry-run',
        action='store_true',
        help='Show what would be renamed without actually doing it'
    )
    
    parser.add_argument(
        '-v', '--verbose',
        action='store_true',
        help='Show detailed output'
    )
    
    parser.add_argument(
        '-r', '--recursive',
        action='store_true',
        help='Process subdirectories recursively'
    )
    
    args = parser.parse_args()
    
    print("=== Remove '.down' Suffix from Filenames ===")
    print()
    
    try:
        total, renamed = process_directory(
            args.directory,
            dry_run=args.dry_run,
            verbose=args.verbose,
            recursive=args.recursive
        )
        
        print()
        print("=== Summary ===")
        print(f"Total files processed: {total}")
        print(f"Files renamed: {renamed}")
        
        if args.dry_run and renamed > 0:
            print()
            print("To actually rename the files, run the command without --dry-run")
        elif renamed > 0:
            print("✅ Operation completed successfully!")
        
    except KeyboardInterrupt:
        print("\n❌ Operation cancelled by user")
        sys.exit(1)
    except Exception as e:
        print(f"❌ Unexpected error: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()


#!/usr/bin/env python3
"""
HamiltonIO SIESTA Command Line Interface

Provides command-line tools for analyzing SIESTA Hamiltonians.
"""

import argparse
import sys
from pathlib import Path


def get_version():
    """Get the HamiltonIO version."""
    try:
        from .. import __version__

        return __version__
    except ImportError:
        return "0.2.5"  # Fallback version from pyproject.toml


def analyze_intra_atomic(
    fdf_file,
    output_file=None,
    atoms=None,
    pauli_decomp=True,
    show_matrix=False,
    read_soc=False,
    ispin=None,
):
    """
    Analyze intra-atomic (on-site) Hamiltonian for SIESTA calculations.

    Parameters:
        fdf_file: str or Path
            Path to SIESTA .fdf file
        output_file: str or None
            Output file path. If None, print to stdout
        atoms: list of int or None
            List of atom indices to analyze. If None, analyze all atoms
        pauli_decomp: bool
            Whether to perform Pauli decomposition for spinor systems
        show_matrix: bool
            Whether to show full matrix elements
        read_soc: bool
            Whether to read split-SOC data (H = H_nosoc + H_soc)
        ispin: str or None
            Spin channel: None, 'up', 'down', or 'merge'

    Returns:
        SiestaHamiltonian: The loaded Hamiltonian object
    """
    from .sisl_wrapper import SislParser
    from HamiltonIO.print_hamiltonian import print_intra_atomic_hamiltonian

    # Validate input
    fdf_path = Path(fdf_file)
    if not fdf_path.exists():
        raise FileNotFoundError(f"FDF file not found: {fdf_path}")

    # Load Hamiltonian
    parser = SislParser(fdf_fname=str(fdf_path), ispin=ispin, read_H_soc=read_soc)
    ham = parser.get_model()

    # Perform analysis
    print_intra_atomic_hamiltonian(
        ham,
        atom_indices=atoms,
        output_file=output_file,
        pauli_decomp=pauli_decomp,
        show_matrix=show_matrix,
    )

    return ham


def intra_atomic_command(args):
    """Handle the intra-atomic analysis command."""
    print("=== HamiltonIO SIESTA Intra-Atomic Analysis ===")
    print(f"Version: {get_version()}")
    print()

    # Validate input file
    fdf_path = Path(args.fdf)
    if not fdf_path.exists():
        print(f"✗ Error: FDF file not found: {fdf_path}")
        sys.exit(1)

    print(f"✓ Input FDF file: {fdf_path.absolute()}")

    # Check for NetCDF file
    nc_file = fdf_path.parent / (fdf_path.stem + ".nc")
    if not nc_file.exists():
        print(f"✗ Error: NetCDF file not found: {nc_file}")
        print(
            "  Please ensure the SIESTA calculation has completed and generated the .nc file."
        )
        sys.exit(1)

    print(f"✓ NetCDF file: {nc_file}")
    print()

    # Parse atom indices
    atoms = None
    if args.atoms:
        try:
            atoms = [int(x) for x in args.atoms.split(",")]
            print(f"  Analyzing atoms: {atoms}")
        except ValueError:
            print(f"✗ Error: Invalid atom indices: {args.atoms}")
            print("  Please provide comma-separated integers (e.g., '0,1,2')")
            sys.exit(1)
    else:
        print("  Analyzing: all atoms")

    # Other options
    print(f"  Split-SOC: {'enabled' if args.split_soc else 'disabled'}")
    print(f"  Pauli decomposition: {'enabled' if args.pauli else 'disabled'}")
    print(f"  Show matrices: {'yes' if args.show_matrix else 'no'}")
    if args.ispin:
        print(f"  Spin channel: {args.ispin}")
    if args.output:
        print(f"  Output file: {args.output}")
    print()

    # Run analysis
    try:
        print("Loading SIESTA Hamiltonian...")
        ham = analyze_intra_atomic(
            fdf_file=args.fdf,
            output_file=args.output,
            atoms=atoms,
            pauli_decomp=args.pauli,
            show_matrix=args.show_matrix,
            read_soc=args.split_soc,
            ispin=args.ispin,
        )

        print()
        print("=" * 80)
        print("✓ Analysis complete!")
        print("=" * 80)
        print()
        print(f"System: {ham._name}")
        print(f"Number of atoms: {len(ham.atoms) if ham.atoms else 'N/A'}")
        print(f"Basis functions: {ham.nbasis}")
        print(f"Split-SOC: {ham.split_soc}")
        print()

        if args.output:
            output_path = Path(args.output)
            if output_path.exists():
                size = output_path.stat().st_size
                if size < 1024:
                    size_str = f"{size} B"
                elif size < 1024**2:
                    size_str = f"{size / 1024:.1f} KB"
                else:
                    size_str = f"{size / (1024**2):.1f} MB"
                print(f"Output written to: {output_path.absolute()}")
                print(f"File size: {size_str}")
            else:
                print(f"⚠ Warning: Output file not found: {output_path}")

    except FileNotFoundError as e:
        print(f"✗ Error: {e}")
        sys.exit(1)
    except Exception as e:
        print(f"✗ Error during analysis: {e}")
        import traceback

        traceback.print_exc()
        sys.exit(1)


def main():
    """Main CLI entry point."""
    parser = argparse.ArgumentParser(
        prog="hamiltonio-siesta",
        description="HamiltonIO SIESTA Hamiltonian analysis tools",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  # Analyze intra-atomic Hamiltonian for all atoms
  %(prog)s intra-atomic siesta.fdf

  # Analyze specific atoms with split-SOC
  %(prog)s intra-atomic siesta.fdf --atoms 0,1,2 --split-soc

  # Save to file with full matrices
  %(prog)s intra-atomic siesta.fdf -o analysis.txt --show-matrix

  # Analyze with spin-orbit coupling (SOC/non-collinear)
  %(prog)s intra-atomic siesta.fdf --ispin merge --pauli

For more information, visit: https://github.com/mailhexu/HamiltonIO
        """,
    )

    parser.add_argument(
        "--version",
        action="version",
        version=f"HamiltonIO SIESTA CLI {get_version()}",
    )

    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Intra-atomic analysis command
    intra_parser = subparsers.add_parser(
        "intra-atomic",
        help="Analyze intra-atomic (on-site) Hamiltonian",
        description="Extract and analyze the on-site (R=0) Hamiltonian blocks for each atom.",
    )
    intra_parser.add_argument("fdf", help="Path to SIESTA .fdf file")
    intra_parser.add_argument(
        "-o", "--output", help="Output file path (default: print to stdout)"
    )
    intra_parser.add_argument(
        "-a",
        "--atoms",
        help="Comma-separated atom indices to analyze (e.g., '0,1,2'). Default: all atoms",
    )
    intra_parser.add_argument(
        "--no-pauli",
        dest="pauli",
        action="store_false",
        default=True,
        help="Disable Pauli decomposition for spinor systems",
    )
    intra_parser.add_argument(
        "--show-matrix",
        action="store_true",
        help="Show full matrix elements (can be large for big systems)",
    )
    intra_parser.add_argument(
        "--split-soc",
        action="store_true",
        help="Read split-SOC data (H = H_nosoc + H_soc). "
        "Requires SIESTA calculation with SOC.split.SR.SO = True",
    )
    intra_parser.add_argument(
        "--ispin",
        choices=["up", "down", "merge"],
        help="Spin channel for collinear calculations. "
        "Use 'merge' for SOC/non-collinear as spinor system",
    )

    # Parse arguments
    args = parser.parse_args()

    # Handle commands
    if args.command == "intra-atomic":
        intra_atomic_command(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Unit tests for intra-atomic Hamiltonian decomposition functionality.

Purpose:
    Test extraction and decomposition of intra-atomic (R=(0,0,0)) Hamiltonian blocks.

How to run:
    uv run python -m pytest tests/test_intra_atomic.py -v
"""

import numpy as np
import pytest


def create_mock_wannier_hamiltonian(natoms=2, norb_per_atom=3):
    """
    Create a mock Wannier90 Hamiltonian for testing.

    Parameters:
        natoms: number of atoms
        norb_per_atom: number of Wannier functions per atom

    Returns:
        WannierHam: Mock Wannier90 Hamiltonian with atoms set
    """
    from ase.atoms import Atoms

    from HamiltonIO.wannier import WannierHam

    # Create mock atoms
    symbols = ["Fe"] * natoms
    positions = np.array([[0, 0, i * 0.5] for i in range(natoms)], dtype=float)
    atoms = Atoms(symbols=symbols, positions=positions, cell=[5, 5, 5])

    nbasis = natoms * norb_per_atom

    # Create mock data dictionary with R=(0,0,0) and some hoppings
    data = {}
    data[(0, 0, 0)] = np.random.rand(nbasis, nbasis) + 1j * np.random.rand(
        nbasis, nbasis
    )
    data[(1, 0, 0)] = np.random.rand(nbasis, nbasis) * 0.1

    # Make R=0 block Hermitian
    data[(0, 0, 0)] = (data[(0, 0, 0)] + data[(0, 0, 0)].conj().T) / 2

    # Create positions for Wannier centers (near atoms)
    positions_wannier = []
    for iatom in range(natoms):
        atom_pos = positions[iatom]
        for iorb in range(norb_per_atom):
            positions_wannier.append(atom_pos + np.random.rand(3) * 0.01)
    positions_wannier = np.array(positions_wannier, dtype=float)

    ham = WannierHam(nbasis=nbasis, data=data, positions=positions_wannier)
    ham.set_atoms(atoms)

    return ham


def create_mock_hamiltonian(natoms=2, norb_per_atom=3, nspin=1, with_soc=False):
    """
    Create a mock LCAOHamiltonian for testing.

    Parameters:
        natoms: number of atoms
        norb_per_atom: number of orbitals per atom
        nspin: 1 for collinear, 2 for spinor
        with_soc: if True, create SOC-split Hamiltonian
    """
    from ase.atoms import Atoms

    from HamiltonIO.lcao_hamiltonian import LCAOHamiltonian

    # Create mock atoms
    symbols = ["Fe"] * natoms
    positions = [[0, 0, i * 3.0] for i in range(natoms)]
    atoms = Atoms(symbols=symbols, positions=positions, cell=[10, 10, 10])

    nbasis = natoms * norb_per_atom * nspin
    nR = 3  # Just a few R vectors including (0,0,0)

    # Create Rlist with (0,0,0) at index 0
    Rlist = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]])

    # Create random HR and SR
    HR = np.random.rand(nR, nbasis, nbasis) + 1j * np.random.rand(nR, nbasis, nbasis)
    SR = np.random.rand(nR, nbasis, nbasis) + 1j * np.random.rand(nR, nbasis, nbasis)

    # Make Hermitian
    for iR in range(nR):
        HR[iR] = (HR[iR] + HR[iR].conj().T) / 2
        SR[iR] = (SR[iR] + SR[iR].conj().T) / 2

    # Create mock orbitals (ABACUS-style for simplicity)
    from dataclasses import dataclass

    @dataclass
    class MockOrb:
        iatom: int
        sym: str

    orbs = []
    for iatom in range(natoms):
        for iorb in range(norb_per_atom * nspin):
            orbs.append(MockOrb(iatom=iatom, sym=f"orb_{iorb}"))

    if with_soc:
        # Split into non-SOC and SOC parts
        HR_nosoc = HR * 0.8  # 80% non-SOC
        HR_soc = HR * 0.2  # 20% SOC
        ham = LCAOHamiltonian(
            HR=None,
            SR=SR,
            Rlist=Rlist,
            nbasis=nbasis,
            atoms=atoms,
            orbs=orbs,
            nspin=nspin,
            HR_nosoc=HR_nosoc,
            HR_soc=HR_soc,
        )
    else:
        ham = LCAOHamiltonian(
            HR=HR,
            SR=SR,
            Rlist=Rlist,
            nbasis=nbasis,
            atoms=atoms,
            orbs=orbs,
            nspin=nspin,
        )

    return ham


def test_orbital_atom_mapping_collinear():
    """Test orbital-to-atom mapping for collinear system."""
    ham = create_mock_hamiltonian(natoms=2, norb_per_atom=3, nspin=1)
    orb_atom_map = ham._get_orbital_atom_mapping()

    assert len(orb_atom_map) == 2  # 2 atoms
    assert len(orb_atom_map[0]) == 3  # 3 orbitals per atom
    assert len(orb_atom_map[1]) == 3
    assert orb_atom_map[0] == [0, 1, 2]
    assert orb_atom_map[1] == [3, 4, 5]


def test_orbital_atom_mapping_spinor():
    """Test orbital-to-atom mapping for spinor system."""
    ham = create_mock_hamiltonian(natoms=2, norb_per_atom=3, nspin=2)
    orb_atom_map = ham._get_orbital_atom_mapping()

    assert len(orb_atom_map) == 2  # 2 atoms
    assert len(orb_atom_map[0]) == 6  # 3 orbitals * 2 spins per atom
    assert len(orb_atom_map[1]) == 6


def test_get_intra_atomic_blocks_collinear():
    """Test extraction of intra-atomic blocks for collinear system."""
    ham = create_mock_hamiltonian(natoms=2, norb_per_atom=3, nspin=1)
    blocks = ham.get_intra_atomic_blocks()

    assert len(blocks) == 2  # 2 atoms
    assert 0 in blocks
    assert 1 in blocks

    # Check shape of block for atom 0
    assert blocks[0]["H_full"].shape == (3, 3)
    assert blocks[0]["H_nosoc"] is None  # No SOC split
    assert blocks[0]["H_soc"] is None
    assert blocks[0]["orbital_indices"] == [0, 1, 2]


def test_get_intra_atomic_blocks_with_soc():
    """Test extraction of intra-atomic blocks with SOC decomposition."""
    ham = create_mock_hamiltonian(natoms=2, norb_per_atom=3, nspin=2, with_soc=True)
    blocks = ham.get_intra_atomic_blocks()

    assert len(blocks) == 2

    # Check SOC decomposition
    assert blocks[0]["H_nosoc"] is not None
    assert blocks[0]["H_soc"] is not None

    # Verify sum: H_full = H_nosoc + H_soc
    H_reconstructed = blocks[0]["H_nosoc"] + blocks[0]["H_soc"]
    assert np.allclose(H_reconstructed, blocks[0]["H_full"], rtol=1e-10)


def test_get_intra_atomic_blocks_filtered():
    """Test extraction with atom filtering."""
    ham = create_mock_hamiltonian(natoms=3, norb_per_atom=2, nspin=1)
    blocks = ham.get_intra_atomic_blocks(atom_indices=[0, 2])

    assert len(blocks) == 2
    assert 0 in blocks
    assert 2 in blocks
    assert 1 not in blocks


def test_print_intra_atomic_hamiltonian_collinear():
    """Test printing functionality for collinear system."""
    import os
    import tempfile

    from HamiltonIO.print_hamiltonian import print_intra_atomic_hamiltonian

    ham = create_mock_hamiltonian(natoms=2, norb_per_atom=2, nspin=1)

    with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".txt") as f:
        output_file = f.name

    try:
        # Print to file
        print_intra_atomic_hamiltonian(
            ham, output_file=output_file, pauli_decomp=False, show_matrix=False
        )

        # Check file was created and has content
        assert os.path.exists(output_file)
        with open(output_file, "r") as f:
            content = f.read()
            assert "Intra-Atomic Hamiltonian Analysis" in content
            assert "Atom 0" in content
            assert "Atom 1" in content
            assert "Fe" in content  # Element symbol
    finally:
        if os.path.exists(output_file):
            os.remove(output_file)


def test_print_intra_atomic_hamiltonian_spinor_with_pauli():
    """Test printing with Pauli decomposition for spinor system."""
    import os
    import tempfile

    from HamiltonIO.print_hamiltonian import print_intra_atomic_hamiltonian

    ham = create_mock_hamiltonian(natoms=1, norb_per_atom=2, nspin=2)

    with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".txt") as f:
        output_file = f.name

    try:
        # Print with Pauli decomposition
        print_intra_atomic_hamiltonian(
            ham, output_file=output_file, pauli_decomp=True, show_matrix=False
        )

        # Check Pauli components are mentioned
        with open(output_file, "r") as f:
            content = f.read()
            assert "Pauli Decomposition" in content
            assert "I (charge) component" in content or "charge" in content.lower()
            assert "spin-x" in content.lower() or "σₓ" in content or "σx" in content
    finally:
        if os.path.exists(output_file):
            os.remove(output_file)


def test_print_intra_atomic_with_soc_split():
    """Test printing with SOC decomposition."""
    import os
    import tempfile

    from HamiltonIO.print_hamiltonian import print_intra_atomic_hamiltonian

    ham = create_mock_hamiltonian(natoms=1, norb_per_atom=2, nspin=2, with_soc=True)

    with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".txt") as f:
        output_file = f.name

    try:
        print_intra_atomic_hamiltonian(
            ham, output_file=output_file, pauli_decomp=True, show_matrix=False
        )

        with open(output_file, "r") as f:
            content = f.read()
            assert "SOC Decomposition" in content
            assert "Non-SOC Part" in content
            assert "SOC Part" in content
            assert "Verification" in content  # Check for sum verification
    finally:
        if os.path.exists(output_file):
            os.remove(output_file)


def test_wannier_orbital_atom_mapping():
    """Test orbital-to-atom mapping for Wannier90 Hamiltonian."""
    ham = create_mock_wannier_hamiltonian(natoms=2, norb_per_atom=3)

    orb_atom_map = ham._get_orbital_atom_mapping()

    # Should have 2 atoms
    assert len(orb_atom_map) == 2

    # Each atom should have 3 orbitals
    assert len(orb_atom_map[0]) == 3
    assert len(orb_atom_map[1]) == 3

    # All orbitals should be assigned
    assigned_orbitals = sum(orb_atom_map.values(), [])
    assert len(assigned_orbitals) == 6  # 2 atoms * 3 orbitals
    assert len(set(assigned_orbitals)) == 6  # All unique


def test_wannier_get_intra_atomic_blocks():
    """Test extraction of intra-atomic blocks for Wannier90."""
    ham = create_mock_wannier_hamiltonian(natoms=2, norb_per_atom=2)

    blocks = ham.get_intra_atomic_blocks()

    # Should have 2 atoms
    assert len(blocks) == 2
    assert 0 in blocks
    assert 1 in blocks

    # Check block for atom 0
    assert blocks[0]["H_full"].shape == (2, 2)
    assert blocks[0]["H_nosoc"] is None  # SOC not supported
    assert blocks[0]["H_soc"] is None
    assert len(blocks[0]["orbital_indices"]) == 2

    # Check block for atom 1
    assert blocks[1]["H_full"].shape == (2, 2)
    assert len(blocks[1]["orbital_indices"]) == 2


def test_wannier_get_intra_atomic_blocks_filtered():
    """Test extraction with atom filtering for Wannier90."""
    ham = create_mock_wannier_hamiltonian(natoms=3, norb_per_atom=2)

    blocks = ham.get_intra_atomic_blocks(atom_indices=[0, 2])

    assert len(blocks) == 2
    assert 0 in blocks
    assert 2 in blocks
    assert 1 not in blocks


def test_wannier_print_intra_atomic():
    """Test printing intra-atomic Hamiltonian for Wannier90."""
    import os
    import tempfile

    from HamiltonIO.print_hamiltonian import print_intra_atomic_hamiltonian

    ham = create_mock_wannier_hamiltonian(natoms=2, norb_per_atom=2)

    with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".txt") as f:
        output_file = f.name

    try:
        # Print to file
        print_intra_atomic_hamiltonian(
            ham, output_file=output_file, pauli_decomp=False, show_matrix=False
        )

        # Check file was created and has content
        assert os.path.exists(output_file)
        with open(output_file, "r") as f:
            content = f.read()
            assert "Intra-Atomic Hamiltonian Analysis" in content
            assert "Atom 0" in content
            assert "Atom 1" in content
            assert "Fe" in content  # Element symbol
            assert "Wannier" in content  # System name
    finally:
        if os.path.exists(output_file):
            os.remove(output_file)


if __name__ == "__main__":
    # Run tests
    pytest.main([__file__, "-v"])



import os
import numpy as np
import copy
from collections import defaultdict
from dataclasses import dataclass


@dataclass
class BaseHamiltonian:
    is_orthogonal: bool = True
    _name: str = "Base Hamiltonian"
    R2kfactor: float = 2 * np.pi
    _nspin: int = 1
    norb: int = 0
    nbasis: int = 0


class Hamiltonian(BaseHamiltonian):
    """
    Abstract class for tight-binding-like Hamiltonian.
    """

    def __init__(
        self,
        R2kfactor=2 * np.pi,
        nspin=1,
        norb=0,
        is_orthogonal=True,
        _name="Generic Hamiltonian",
    ):
        #: :math:`\alpha` used in :math:`H(k)=\sum_R  H(R) \exp( \alpha k \cdot R)`,
        #: Should be :math:`2\pi i` or :math:`-2\pi i`
        self.is_orthogonal = is_orthogonal

        self.R2kfactor = R2kfactor

        #: number of spin. 1 for collinear, 2 for spinor.
        self._nspin = nspin

        #:number of orbitals. Each orbital can have two spins.
        self.norb = norb

        #: nbasis=nspin*norb
        self.nbasis = nspin * norb

        #: The array of cartesian coordinate of all basis. shape:nbasis,3
        self.xcart = None

        #: The array of cartesian coordinate of all basis. shape:nbasis,3
        self.xred = None

        #: The order of the spinor basis.
        #: 1: orb1_up, orb2_up,  ... orb1_down, orb2_down,...
        #: 2: orb1_up, orb1_down, orb2_up, orb2_down,...
        self._name = "Generic Hamiltonian"

    @property
    def Rlist(self):
        """
        list of R vectors.
        """
        return np.asarray(self._Rlist)

    @Rlist.setter
    def Rlist(self, value):
        self._Rlist = value

    @property
    def nR(self):
        """
        number of R vectors.
        """
        return len(self._Rlist)

    @property
    def orb_names(self):
        """
        list of orbital names.
        """
        return self._orb_names

    @property
    def orb_idx(self):
        """
        dictionary of orbital index, the keys are the orbital names, the keys are the indices.
        """
        return self._orb_idx

    @property
    def name(self):
        """
        name of the Hamiltonian. e.g. "Wannier90", "SIESTA"
        """
        return self._name

    def get_HR(self, R=None, iR=None, ispin=0, dense=True):
        """
        get the Hamiltonian H(R), array of shape (nbasis, nbasis)
        parameters:
        =================
        R: array-like
            the R vector.
        iR: int
            the index of the R vector.
        ispin: int
            the index of the spin. For collinear system: 0 for up, 1 for down, None for both.
              For non-spin-polarized system, or non-collinear/SOC system, ispin is ignored.
        dense: bool
            if True, convert to dense matrix if the default format is sparse.

        Returns:
        =================
        H: array of shape (nbasis, nbasis)
        """
        raise NotImplementedError()

    def get_all_HR(self, dense=True, ispin=None):
        """
        get the Hamiltonian H(R) for all R vectors.
        parameters:
        =================
        dense: bool
            if True, convert to dense matrix if the default format is sparse.
        ispin: int or None
            For collinear system: 0 for up, 1 for down, None for both.
            For others, ispin is ignored.
        Returns:
        =================
        H:  array of shape ( nR, nbasis, nbasis)
          if ispin is None, and kspin=2, H.shape=(nR, nbasis, nbasis, 2)

        """
        raise NotImplementedError()

    def get_orbs(self):
        """
        returns the orbitals.
        """
        raise NotImplementedError()

    def HSE_k(self, kpt):
        raise NotImplementedError()

    def HS_and_eigen(self, kpts):
        """
        get Hamiltonian, overlap matrices, eigenvalues, eigen vectors for all kpoints.

        :param:

        * kpts: list of k points.

        :returns:

        * H, S, eigenvalues, eigenvectors for all kpoints
        * H: complex array of shape (nkpts, nbasis, nbasis)
        * S: complex array of shape (nkpts, nbasis, nbasis). S=None if the basis set is orthonormal.
        * evals: complex array of shape (nkpts, nbands)
        * evecs: complex array of shape (nkpts, nbasis, nbands)
        """
        raise NotImplementedError()

    def get_Hk(self, kpt, ispin=None, dense=True):
        """
        get the Hamiltonian H(k), array of shape (nbasis, nbasis)
        parameters:
        =================
        kpt: array-like
            the k vector.
        ispin: int
            the index of the spin. For collinear system: 0 for up, 1 for down, None for both.
              For non-spin-polarized system, or non-collinear/SOC system, ispin is ignored.
        dense: bool
            if True, convert to dense matrix if the default format is sparse.

        Returns:
        =================
        H: array of shape (nbasis, nbasis)
        """
        raise NotImplementedError()

    def get_Sk(self, kpt, ispin=None, dense=True):
        """
        get the overlap matrix S(k), array of shape (nbasis, nbasis)
        parameters:
        =================
        kpt: array-like
            the k vector.
        ispin: int
            the index of the spin. For collinear system: 0 for up, 1 for down, None for both.
              For non-spin-polarized system, or non-collinear/SOC system, ispin is ignored.
        dense: bool
            if True, convert to dense matrix if the default format is sparse.

        Returns:
        =================
        S: array of shape (nbasis, nbasis)
        """
        raise NotImplementedError()

    def get_HS_and_eigen_k(self, kpt, ispin=None, dense=True):
        """
        get Hamiltonian, overlap matrices, eigenvalues, eigen vectors for a kpoint.
        parameters:
        =================
        kpt: array-like
            the k vector.
        ispin: int
            the index of the spin. For collinear system: 0 for up, 1 for down, None for both.
              For non-spin-polarized system, or non-collinear/SOC system, ispin is ignored.
        dense: bool
            if True, convert to dense matrix if the default format is sparse.

        returns:
        =================
        H, S, eigenvalues, eigenvectors
        H: array of shape (nbasis, nbasis)
        S: array of shape (nbasis, nbasis). S=None if the basis set is orthonormal.
        evals: array of shape (nbands,)
        evecs: array of shape (nbasis, nbands)
        """
        raise NotImplementedError()



import numpy as np
from HamiltonIO.occupations import Occupation, get_occupation


def get_density_matrix_k(
    evals=None, evecs=None, kweights=None, nel=None, width=0.1, S=None
):
    """
    Compute the density matrix in k-space.
    """
    occ = get_occupation(evals, kweights, nel, width=width)
    rho = np.einsum("kib, kb, kjb -> kij", evecs, occ, evecs.conj())
    for ik in range(len(rho)):
        rho[ik] = rho[ik] @ S[ik]
    return rho


def get_total_density_matrix(rho, kweights):
    """
    Compute the total density matrix."""
    return np.einsum("kij, k -> ij", rho, kweights)



import sisl
import numpy as np
import matplotlib.pyplot as plt
import os
import sys


############################################################################
def calctype(data):
    """
     decide the type of calculation
    """
    if data == "unpolarized":
        outdata = "NP"

    elif data == "polarized":
        outdata = "SP"

    elif data == "non-colinear":
        outdata = "SOC"

    elif data == "spin-orbit":
        outdata = "SOC"

    return outdata


def ncases(var):
    return "%8f" % var


###############################################################################


def output_siesta_wantibexos(inputfdf = "MoS2_SZP_SOC/RUN.fdf"):
    tshs = sisl.get_sile(inputfdf).read_hamiltonian()  # pegar hamiltoniano do siesta
    fermi = sisl.get_sile(inputfdf).read_fermi_level()
    scs = 0.0000  # scissors operator
    nbasis = tshs.no
    ncell = tshs.nsc[0] * tshs.nsc[1] * tshs.nsc[2]
    sptype = str(tshs.spin)
    sptype2 = sptype[5:-9]

    if sptype2 == "unpolarized":
        output_unpolarized()

def output_unpolarized(info_fname="system-info-NP.txt", output_fname="siesta_tb-NP.dat"):
        f = open(output_fname, "a")
        print(tshs, file=f)
        f.close()
    
        nbasis = tshs.no
        ncell = tshs.nsc[0] * tshs.nsc[1] * tshs.nsc[2]
        sptype = str(tshs.spin)
        f = open("siesta_tb-NP.dat", "a")
    
        print(calctype(sptype2), file=f)
        print(ncases(scs), file=f)
        print(ncases(fermi[0]), file=f)
        print(
            ncases(tshs.cell[0, 0]),
            "",
            ncases(tshs.cell[0, 1]),
            "",
            ncases(tshs.cell[0, 2]),
            file=f,
        )
        print(
            ncases(tshs.cell[1, 0]),
            "",
            ncases(tshs.cell[1, 1]),
            "",
            ncases(tshs.cell[1, 2]),
            file=f,
        )
        print(
            ncases(tshs.cell[2, 0]),
            "",
            ncases(tshs.cell[2, 1]),
            "",
            ncases(tshs.cell[2, 2]),
            file=f,
        )
        print(nbasis, file=f)
        print(ncell, file=f)
        print(
            "#rcell x",
            " ",
            "rcell y",
            " ",
            "rcell z",
            " ",
            "i",
            " ",
            "j",
            " ",
            "ReH",
            " ",
            "ImH",
            " ",
            "S",
            file=f,
        )
    
        for i in range(0, ncell):
            for j in range(0, nbasis):
                for k in range(0, nbasis):
                    a = tshs.geometry.o2sc(k + i * nbasis)[0]
                    b = tshs.geometry.o2sc(k + i * nbasis)[1]
                    c = tshs.geometry.o2sc(k + i * nbasis)[2]
    
                    d = tshs.H[j, k + i * nbasis]
                    e = tshs.S[j, k + i * nbasis]
    
                    print(
                        ncases(a),
                        " ",
                        ncases(b),
                        " ",
                        ncases(c),
                        " ",
                        j + 1,
                        " ",
                        k + 1,
                        " ",
                        ncases(d),
                        " ",
                        ncases(0.00),
                        " ",
                        ncases(e),
                        file=f,
                    )
    
        f.close()
    
        f = open("siesta_basis_set-NP", "a")
    
        print("bindex aspecie ax ay az l m spin", file=f)
    
        io = 0
        for ia, atom in enumerate(tshs.geometry.atoms):
            xyz = tshs.geometry.xyz[ia, :]
            for orbital in atom:
                print(
                    io + 1,
                    " ",
                    atom.tag,
                    " ",
                    ncases(xyz[0]),
                    " ",
                    ncases(xyz[1]),
                    " ",
                    ncases(xyz[2]),
                    " ",
                    orbital.l,
                    " ",
                    orbital.m,
                    " ",
                    0,
                    file=f,
                )
                io += 1
    
        f.close()
    
    
    ####################################################################
    
    if sptype2 == "polarized":
        f = open("system-info-sp.txt", "a")
        print(tshs, file=f)
        f.close()
    
        nbasis = tshs.no
        ncell = tshs.nsc[0] * tshs.nsc[1] * tshs.nsc[2]
        sptype = str(tshs.spin)
    
        f = open("siesta_basis_set-sp", "a")
    
        print("#bindex aspecie ax ay az l m spin", file=f)
    
        io = 0
        for ia, atom in enumerate(tshs.geometry.atoms):
            xyz = tshs.geometry.xyz[ia, :]
            for orbital in atom:
                print(
                    io + 1,
                    " ",
                    atom.tag,
                    " ",
                    ncases(xyz[0]),
                    " ",
                    ncases(xyz[1]),
                    " ",
                    ncases(xyz[2]),
                    " ",
                    orbital.l,
                    " ",
                    orbital.m,
                    " ",
                    1,
                    file=f,
                )
                io += 1
    
        io = 0
        for ia, atom in enumerate(tshs.geometry.atoms):
            xyz = tshs.geometry.xyz[ia, :]
            for orbital in atom:
                print(
                    nbasis + io + 1,
                    " ",
                    atom.tag,
                    " ",
                    ncases(xyz[0]),
                    " ",
                    ncases(xyz[1]),
                    " ",
                    ncases(xyz[2]),
                    " ",
                    orbital.l,
                    " ",
                    orbital.m,
                    " ",
                    -1,
                    file=f,
                )
                io += 1
    
        f.close()
    
        f = open("siesta_tb-sp.dat", "a")
    
        print(calctype(sptype2), file=f)
        print(ncases(scs), file=f)
        print(ncases(fermi[0]), file=f)
        print(
            ncases(tshs.cell[0, 0]),
            "",
            ncases(tshs.cell[0, 1]),
            "",
            ncases(tshs.cell[0, 2]),
            file=f,
        )
        print(
            ncases(tshs.cell[1, 0]),
            "",
            ncases(tshs.cell[1, 1]),
            "",
            ncases(tshs.cell[1, 2]),
            file=f,
        )
        print(
            ncases(tshs.cell[2, 0]),
            "",
            ncases(tshs.cell[2, 1]),
            "",
            ncases(tshs.cell[2, 2]),
            file=f,
        )
        print(2 * nbasis, file=f)
        print(ncell, file=f)
        print(
            "#rcell x",
            " ",
            "rcell y",
            " ",
            "rcell z",
            " ",
            "i",
            " ",
            "j",
            " ",
            "ReH",
            " ",
            "ImH",
            " ",
            "S",
            file=f,
        )
        # np.empty aloca todos os arrays vazios
        # alocando todos os arrays com 0
    
        S = np.zeros((ncell + 1, (2 * nbasis) + 2, (2 * nbasis) + 2))
        reH = np.zeros((ncell + 1, (2 * nbasis) + 2, (2 * nbasis) + 2))
        imH = np.zeros((ncell + 1, (2 * nbasis) + 2, (2 * nbasis) + 2))
    
        a = np.zeros((ncell + 1, (2 * nbasis) + 2))
        b = np.zeros((ncell + 1, (2 * nbasis) + 2))
        c = np.zeros((ncell + 1, (2 * nbasis) + 2))
    
        for i in range(0, ncell):
            for j in range(0, nbasis):
                for k in range(0, nbasis):
                    # parte up-up
    
                    reH[i + 1, j + 1, k + 1] = tshs[j, k + i * nbasis][0]
                    # imH[i+1,j+1,k+1] = tshs[j,k+i*nbasis][4]
                    S[i + 1, j + 1, k + 1] = tshs.S[j, k + i * nbasis]
    
                    # parte up-dn
    
                    # reH[i+1,(j+1),(k+1)+nbasis] = tshs[j,k+i*nbasis][2]
                    # imH[i+1,(j+1),(k+1)+nbasis] = tshs[j,k+i*nbasis][3]
                    # S[i+1,(j+1),(k+1)+nbasis] = 0.0
    
                    # parte dn-up
    
                    # reH[i+1,(j+1)+nbasis,(k+1)] = tshs[j,k+i*nbasis][6]
                    # imH[i+1,(j+1)+nbasis,(k+1)] = tshs[j,k+i*nbasis][7]
                    # S[i+1,(j+1)+nbasis,(k+1)] = 0.0
    
                    # parte dn-dn
    
                    reH[i + 1, (j + 1) + nbasis, (k + 1) + nbasis] = tshs[
                        j, k + i * nbasis
                    ][1]
                    # imH[i+1,(j+1)+nbasis,(k+1)+nbasis] = tshs[j,k+i*nbasis][5]
                    S[i + 1, (j + 1) + nbasis, (k + 1) + nbasis] = tshs.S[j, k + i * nbasis]
    
        # escrevendo a localização dos atomos da base
    
        for i in range(0, ncell):
            for k in range(0, nbasis):
                a[i + 1, k + 1] = tshs.geometry.o2sc(k + i * nbasis)[0]
                b[i + 1, k + 1] = tshs.geometry.o2sc(k + i * nbasis)[1]
                c[i + 1, k + 1] = tshs.geometry.o2sc(k + i * nbasis)[2]
    
                a[i + 1, (k + 1) + nbasis] = tshs.geometry.o2sc(k + i * nbasis)[0]
                b[i + 1, (k + 1) + nbasis] = tshs.geometry.o2sc(k + i * nbasis)[1]
                c[i + 1, (k + 1) + nbasis] = tshs.geometry.o2sc(k + i * nbasis)[2]
    
        for i in range(1, ncell + 1):
            for j in range(1, (2 * nbasis) + 1):
                for k in range(1, (2 * nbasis) + 1):
                    #   a = tshs.geometry.o2sc(k+i*nbasis)[0]
                    #   b = tshs.geometry.o2sc(k+i*nbasis)[1]
                    #   c = tshs.geometry.o2sc(k+i*nbasis)[2]
    
                    #   d = tshs.H[j,k+i*nbasis]
                    #   e = tshs.S[j,k+i*nbasis]
    
                    print(
                        ncases(a[i, k]),
                        " ",
                        ncases(b[i, k]),
                        " ",
                        ncases(c[i, k]),
                        " ",
                        j,
                        " ",
                        k,
                        " ",
                        ncases(reH[i, j, k]),
                        " ",
                        ncases(imH[i, j, k]),
                        " ",
                        ncases(S[i, j, k]),
                        file=f,
                    )
    
        f.close()
    
    ####################################################################
    
    if sptype2 == "non-colinear":
        f = open("system-info-nc.txt", "a")
        print(tshs, file=f)
        f.close()
    
        nbasis = tshs.no
        ncell = tshs.nsc[0] * tshs.nsc[1] * tshs.nsc[2]
        sptype = str(tshs.spin)
    
        f = open("siesta_basis_set-nc", "a")
    
        print("#bindex aspecie ax ay az l m spin", file=f)
    
        io = 0
        for ia, atom in enumerate(tshs.geometry.atoms):
            xyz = tshs.geometry.xyz[ia, :]
            for orbital in atom:
                print(
                    io + 1,
                    " ",
                    atom.tag,
                    " ",
                    ncases(xyz[0]),
                    " ",
                    ncases(xyz[1]),
                    " ",
                    ncases(xyz[2]),
                    " ",
                    orbital.l,
                    " ",
                    orbital.m,
                    " ",
                    1,
                    file=f,
                )
                io += 1
    
        io = 0
        for ia, atom in enumerate(tshs.geometry.atoms):
            xyz = tshs.geometry.xyz[ia, :]
            for orbital in atom:
                print(
                    nbasis + io + 1,
                    " ",
                    atom.tag,
                    " ",
                    ncases(xyz[0]),
                    " ",
                    ncases(xyz[1]),
                    " ",
                    ncases(xyz[2]),
                    " ",
                    orbital.l,
                    " ",
                    orbital.m,
                    " ",
                    -1,
                    file=f,
                )
                io += 1
    
        f.close()
    
        f = open("siesta_tb-nc.dat", "a")
    
        print(calctype(sptype2), file=f)
        print(ncases(scs), file=f)
        print(ncases(fermi[0]), file=f)
        print(
            ncases(tshs.cell[0, 0]),
            "",
            ncases(tshs.cell[0, 1]),
            "",
            ncases(tshs.cell[0, 2]),
            file=f,
        )
        print(
            ncases(tshs.cell[1, 0]),
            "",
            ncases(tshs.cell[1, 1]),
            "",
            ncases(tshs.cell[1, 2]),
            file=f,
        )
        print(
            ncases(tshs.cell[2, 0]),
            "",
            ncases(tshs.cell[2, 1]),
            "",
            ncases(tshs.cell[2, 2]),
            file=f,
        )
        print(2 * nbasis, file=f)
        print(ncell, file=f)
        print(
            "#rcell x",
            " ",
            "rcell y",
            " ",
            "rcell z",
            " ",
            "i",
            " ",
            "j",
            " ",
            "ReH",
            " ",
            "ImH",
            " ",
            "S",
            file=f,
        )
    
        # np.empty aloca todos os arrays vazios
        # alocando todos os arrays com 0
    
        S = np.zeros((ncell + 1, (2 * nbasis) + 2, (2 * nbasis) + 2))
        reH = np.zeros((ncell + 1, (2 * nbasis) + 2, (2 * nbasis) + 2))
        imH = np.zeros((ncell + 1, (2 * nbasis) + 2, (2 * nbasis) + 2))
    
        a = np.zeros((ncell + 1, (2 * nbasis) + 2))
        b = np.zeros((ncell + 1, (2 * nbasis) + 2))
        c = np.zeros((ncell + 1, (2 * nbasis) + 2))
    
        for i in range(0, ncell):
            for j in range(0, nbasis):
                for k in range(0, nbasis):
                    # parte up-up
    
                    reH[i + 1, j + 1, k + 1] = tshs[j, k + i * nbasis][0]
                    # imH[i+1,j+1,k+1] = tshs[j,k+i*nbasis][4]
                    S[i + 1, j + 1, k + 1] = tshs.S[j, k + i * nbasis]
    
                    # parte up-dn
    
                    reH[i + 1, (j + 1), (k + 1) + nbasis] = tshs[j, k + i * nbasis][2]
                    # imH[i+1,(j+1),(k+1)+nbasis] = tshs[j,k+i*nbasis][3]
                    S[i + 1, (j + 1), (k + 1) + nbasis] = 0.0
    
                    # parte dn-up
    
                    reH[i + 1, (j + 1) + nbasis, (k + 1)] = tshs[j, k + i * nbasis][3]
                    # imH[i+1,(j+1)+nbasis,(k+1)] = tshs[j,k+i*nbasis][7]
                    S[i + 1, (j + 1) + nbasis, (k + 1)] = 0.0
    
                    # parte dn-dn
    
                    reH[i + 1, (j + 1) + nbasis, (k + 1) + nbasis] = tshs[
                        j, k + i * nbasis
                    ][1]
                    # imH[i+1,(j+1)+nbasis,(k+1)+nbasis] = tshs[j,k+i*nbasis][5]
                    S[i + 1, (j + 1) + nbasis, (k + 1) + nbasis] = tshs.S[j, k + i * nbasis]
    
        # escrevendo a localização dos atomos da base
    
        for i in range(0, ncell):
            for k in range(0, nbasis):
                a[i + 1, k + 1] = tshs.geometry.o2sc(k + i * nbasis)[0]
                b[i + 1, k + 1] = tshs.geometry.o2sc(k + i * nbasis)[1]
                c[i + 1, k + 1] = tshs.geometry.o2sc(k + i * nbasis)[2]
    
                a[i + 1, (k + 1) + nbasis] = tshs.geometry.o2sc(k + i * nbasis)[0]
                b[i + 1, (k + 1) + nbasis] = tshs.geometry.o2sc(k + i * nbasis)[1]
                c[i + 1, (k + 1) + nbasis] = tshs.geometry.o2sc(k + i * nbasis)[2]
    
        for i in range(1, ncell + 1):
            for j in range(1, (2 * nbasis) + 1):
                for k in range(1, (2 * nbasis) + 1):
                    #   a = tshs.geometry.o2sc(k+i*nbasis)[0]
                    #   b = tshs.geometry.o2sc(k+i*nbasis)[1]
                    #   c = tshs.geometry.o2sc(k+i*nbasis)[2]
    
                    #   d = tshs.H[j,k+i*nbasis]
                    #   e = tshs.S[j,k+i*nbasis]
    
                    print(
                        ncases(a[i, k]),
                        " ",
                        ncases(b[i, k]),
                        " ",
                        ncases(c[i, k]),
                        " ",
                        j,
                        " ",
                        k,
                        " ",
                        ncases(reH[i, j, k]),
                        " ",
                        ncases(imH[i, j, k]),
                        " ",
                        ncases(S[i, j, k]),
                        file=f,
                    )
    
        f.close()
    
    ####################################################################
    
    if sptype2 == "spin-orbit":
        f = open("siesta_basis_set-soc", "a")
    
        print("#bindex aspecie ax ay az l m spin", file=f)
    
        io = 0
        for ia, atom in enumerate(tshs.geometry.atoms):
            xyz = tshs.geometry.xyz[ia, :]
            for orbital in atom:
                print(
                    io + 1,
                    " ",
                    atom.tag,
                    " ",
                    ncases(xyz[0]),
                    " ",
                    ncases(xyz[1]),
                    " ",
                    ncases(xyz[2]),
                    " ",
                    orbital.l,
                    " ",
                    orbital.m,
                    " ",
                    1,
                    file=f,
                )
                io += 1
    
        io = 0
        for ia, atom in enumerate(tshs.geometry.atoms):
            xyz = tshs.geometry.xyz[ia, :]
            for orbital in atom:
                print(
                    nbasis + io + 1,
                    " ",
                    atom.tag,
                    " ",
                    ncases(xyz[0]),
                    " ",
                    ncases(xyz[1]),
                    " ",
                    ncases(xyz[2]),
                    " ",
                    orbital.l,
                    " ",
                    orbital.m,
                    " ",
                    -1,
                    file=f,
                )
                io += 1
    
        f.close()
    
        f = open("siesta_tb-soc.dat", "a")
    
        print(calctype(sptype2), file=f)
        print(ncases(scs), file=f)
        print(ncases(fermi[0]), file=f)
        print(
            ncases(tshs.cell[0, 0]),
            "",
            ncases(tshs.cell[0, 1]),
            "",
            ncases(tshs.cell[0, 2]),
            file=f,
        )
        print(
            ncases(tshs.cell[1, 0]),
            "",
            ncases(tshs.cell[1, 1]),
            "",
            ncases(tshs.cell[1, 2]),
            file=f,
        )
        print(
            ncases(tshs.cell[2, 0]),
            "",
            ncases(tshs.cell[2, 1]),
            "",
            ncases(tshs.cell[2, 2]),
            file=f,
        )
        print(2 * nbasis, file=f)
        print(ncell, file=f)
        print(
            "#rcell x",
            " ",
            "rcell y",
            " ",
            "rcell z",
            " ",
            "i",
            " ",
            "j",
            " ",
            "ReH",
            " ",
            "ImH",
            " ",
            "S",
            file=f,
        )
    
        # np.empty aloca todos os arrays vazios
        # alocando todos os arrays com 0
    
        S = np.zeros((ncell + 1, (2 * nbasis) + 2, (2 * nbasis) + 2))
        reH = np.zeros((ncell + 1, (2 * nbasis) + 2, (2 * nbasis) + 2))
        imH = np.zeros((ncell + 1, (2 * nbasis) + 2, (2 * nbasis) + 2))
    
        a = np.zeros((ncell + 1, (2 * nbasis) + 2))
        b = np.zeros((ncell + 1, (2 * nbasis) + 2))
        c = np.zeros((ncell + 1, (2 * nbasis) + 2))
    
        for i in range(0, ncell):
            for j in range(0, nbasis):
                for k in range(0, nbasis):
                    # parte up-up
    
                    reH[i + 1, j + 1, k + 1] = tshs[j, k + i * nbasis][0]
                    imH[i + 1, j + 1, k + 1] = tshs[j, k + i * nbasis][4]
                    S[i + 1, j + 1, k + 1] = tshs.S[j, k + i * nbasis]
    
                    # parte up-dn
    
                    reH[i + 1, (j + 1), (k + 1) + nbasis] = tshs[j, k + i * nbasis][2]
                    imH[i + 1, (j + 1), (k + 1) + nbasis] = tshs[j, k + i * nbasis][3]
                    S[i + 1, (j + 1), (k + 1) + nbasis] = 0.0
    
                    # parte dn-up
    
                    reH[i + 1, (j + 1) + nbasis, (k + 1)] = tshs[j, k + i * nbasis][6]
                    imH[i + 1, (j + 1) + nbasis, (k + 1)] = tshs[j, k + i * nbasis][7]
                    S[i + 1, (j + 1) + nbasis, (k + 1)] = 0.0
    
                    # parte dn-dn
    
                    reH[i + 1, (j + 1) + nbasis, (k + 1) + nbasis] = tshs[
                        j, k + i * nbasis
                    ][1]
                    imH[i + 1, (j + 1) + nbasis, (k + 1) + nbasis] = tshs[
                        j, k + i * nbasis
                    ][5]
                    S[i + 1, (j + 1) + nbasis, (k + 1) + nbasis] = tshs.S[j, k + i * nbasis]
    
        # escrevendo a localização dos atomos da base
    
        for i in range(0, ncell):
            for k in range(0, nbasis):
                a[i + 1, k + 1] = tshs.geometry.o2sc(k + i * nbasis)[0]
                b[i + 1, k + 1] = tshs.geometry.o2sc(k + i * nbasis)[1]
                c[i + 1, k + 1] = tshs.geometry.o2sc(k + i * nbasis)[2]
    
                a[i + 1, (k + 1) + nbasis] = tshs.geometry.o2sc(k + i * nbasis)[0]
                b[i + 1, (k + 1) + nbasis] = tshs.geometry.o2sc(k + i * nbasis)[1]
                c[i + 1, (k + 1) + nbasis] = tshs.geometry.o2sc(k + i * nbasis)[2]
    
        for i in range(1, ncell + 1):
            for j in range(1, (2 * nbasis) + 1):
                for k in range(1, (2 * nbasis) + 1):
                    #   a = tshs.geometry.o2sc(k+i*nbasis)[0]
                    #   b = tshs.geometry.o2sc(k+i*nbasis)[1]
                    #   c = tshs.geometry.o2sc(k+i*nbasis)[2]
    
                    #   d = tshs.H[j,k+i*nbasis]
                    #   e = tshs.S[j,k+i*nbasis]
    
                    print(
                        ncases(a[i, k]),
                        " ",
                        (b[i, k]),
                        " ",
                        ncases(c[i, k]),
                        " ",
                        j,
                        " ",
                        k,
                        " ",
                        ncases(reH[i, j, k]),
                        " ",
                        ncases(imH[i, j, k]),
                        " ",
                        ncases(S[i, j, k]),
                        file=f,
                    )
    
        f.close()
    

####################################################################
